# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from typing import Sequence, Dict, Optional, Mapping, List, Tuple
import os

from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.var import VarData
from pyomo.core.base.param import ParamData
from pyomo.core.base.block import BlockData
from pyomo.core.base.objective import Objective, ObjectiveData
from pyomo.common.config import ConfigValue, ConfigDict
from pyomo.common.enums import IntEnum, SolverAPIVersion
from pyomo.common.errors import ApplicationError
from pyomo.common.deprecation import deprecation_warning
from pyomo.common.modeling import NOTSET
from pyomo.opt.results.results_ import SolverResults as LegacySolverResults
from pyomo.opt.results.solution import Solution as LegacySolution
from pyomo.core.kernel.objective import minimize
from pyomo.core.base import SymbolMap
from pyomo.core.base.label import NumericLabeler
from pyomo.core.staleflag import StaleFlagManager
from pyomo.scripting.solve_config import default_config_block
from pyomo.contrib.solver.common.config import SolverConfig, PersistentSolverConfig
from pyomo.contrib.solver.common.util import get_objective
from pyomo.contrib.solver.common.results import (
    Results,
    legacy_solver_status_map,
    legacy_termination_condition_map,
    legacy_solution_status_map,
)


class Availability(IntEnum):
    """
    Class to capture different statuses in which a solver can exist in
    order to record its availability for use.
    """

    FullLicense = 2
    LimitedLicense = 1
    NotFound = 0
    BadVersion = -1
    BadLicense = -2
    NeedsCompiledExtension = -3

    def __bool__(self):
        return self._value_ > 0

    def __format__(self, format_spec):
        return format(self.name, format_spec)

    def __str__(self):
        return self.name


class SolverBase:
    """The base class for "new-style" Pyomo solver interfaces.

    This base class defines the methods all derived solvers are expected
    to implement:

      - :py:meth:`available`
      - :py:meth:`is_persistent`
      - :py:meth:`solve`
      - :py:meth:`version`

    **Class Configuration**

    All derived concrete implementations of this class must define a
    class attribute ``CONFIG`` containing the :class:`ConfigDict` that
    specifies the solver's configuration options.  By convention, the
    ``CONFIG`` should derive from (or implement a superset of the
    options from) one of the following:

    - :class:`~pyomo.contrib.solver.common.config.SolverConfig`
    - :class:`~pyomo.contrib.solver.common.config.BranchAndBoundConfig`
    - :class:`~pyomo.contrib.solver.common.config.PersistentSolverConfig`
    - :class:`~pyomo.contrib.solver.common.config.PersistentBranchAndBoundConfig`

    """

    CONFIG = SolverConfig()

    def __init__(self, **kwds) -> None:
        if "name" in kwds:
            self.name = kwds.pop('name')
        elif not hasattr(self, 'name'):
            self.name = type(self).__name__.lower()

        #: Instance configuration; see CONFIG documentation on derived class
        self.config = self.CONFIG(value=kwds)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exit statement - enables `with` statements."""

    @classmethod
    def api_version(self):
        """
        Return the public API supported by this interface.

        Returns
        -------
        ~pyomo.common.enums.SolverAPIVersion
            A solver API enum object
        """
        return SolverAPIVersion.V2

    def solve(self, model: BlockData, **kwargs) -> Results:
        """Solve a Pyomo model.

        Parameters
        ----------
        model: BlockData
            The Pyomo model to be solved
        **kwargs
            Additional keyword arguments (including solver_options - passthrough
            options; delivered directly to the solver (with no validation))

        Returns
        -------
        results: :class:`Results<pyomo.contrib.solver.common.results.Results>`
            A results object

        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'solve'."
        )

    def available(self) -> Availability:
        """Test if the solver is available on this system.

        Nominally, this will return `True` if the solver interface is
        valid and can be used to solve problems and `False` if it cannot.
        Note that for licensed solvers there are a number of "levels" of
        available: depending on the license, the solver may be available
        with limitations on problem size or runtime (e.g., 'demo'
        vs. 'community' vs. 'full').  In these cases, the solver may
        return a subclass of enum.IntEnum, with members that resolve to
        True if the solver is available (possibly with limitations).
        The Enum may also have multiple members that all resolve to
        False indicating the reason why the interface is not available
        (not found, bad license, unsupported version, etc).

        Returns
        -------
        available: Availability
            An enum that indicates "how available" the solver is.
            Note that the enum can be cast to bool, which will
            be True if the solver is runnable at all and False
            otherwise.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'available'."
        )

    def version(self) -> Tuple:
        """Return the solver version found on the system.

        Returns
        -------
        version: tuple
            A tuple representing the version
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'version'."
        )

    def is_persistent(self) -> bool:
        """``True`` if this supports a persistent interface to the solver.

        Returns
        -------
        is_persistent: bool
            True if the solver is a persistent solver.
        """
        return False


class PersistentSolverBase(SolverBase):
    """
    Base class upon which persistent solvers can be built. This inherits the
    methods from the solver base class and adds those methods that are necessary
    for persistent solvers.

    Example usage can be seen in the Gurobi interface.
    """

    CONFIG = PersistentSolverConfig()

    def __init__(self, **kwds) -> None:
        super().__init__(**kwds)
        self._active_config = self.config

    def solve(self, model: BlockData, **kwargs) -> Results:
        """
        Solve a Pyomo model.

        Parameters
        ----------
        model: BlockData
            The Pyomo model to be solved
        **kwargs
            Additional keyword arguments (including solver_options - passthrough
            options; delivered directly to the solver (with no validation))

        Returns
        -------
        results: :class:`Results<pyomo.contrib.solver.common.results.Results>`
            A results object
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'solve'."
        )

    def is_persistent(self) -> bool:
        """
        Returns
        -------
        is_persistent: bool
            True if the solver is a persistent solver.
        """
        return True

    def _load_vars(self, vars_to_load: Optional[Sequence[VarData]] = None) -> None:
        """
        Load the solution of the primal variables into the value attribute of the variables.

        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose solution should be loaded. If vars_to_load
            is None, then the solution to all primal variables will be loaded.
        """
        for var, val in self._get_primals(vars_to_load=vars_to_load).items():
            var.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    def _get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        """
        Get mapping of variables to primals.

        Parameters
        ----------
        vars_to_load : Optional[Sequence[VarData]], optional
            Which vars to be populated into the map. The default is None.

        Returns
        -------
        Mapping[VarData, float]
            A map of variables to primals.
        """
        raise NotImplementedError(
            f'{type(self)} does not support the get_primals method'
        )

    def _get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        """
        Declare sign convention in docstring here.

        Parameters
        ----------
        cons_to_load: list
            A list of the constraints whose duals should be loaded. If cons_to_load
            is None, then the duals for all constraints will be loaded.

        Returns
        -------
        duals: dict
            Maps constraints to dual values
        """
        raise NotImplementedError(f'{type(self)} does not support the get_duals method')

    def _get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        """
        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose reduced cost should be loaded. If vars_to_load
            is None, then all reduced costs will be loaded.

        Returns
        -------
        reduced_costs: ComponentMap
            Maps variable to reduced cost
        """
        raise NotImplementedError(
            f'{type(self)} does not support the get_reduced_costs method'
        )

    def set_instance(self, model: BlockData):
        """
        Set an instance of the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'set_instance'."
        )

    def set_objective(self, obj: ObjectiveData):
        """
        Set current objective for the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'set_objective'."
        )

    def add_constraints(self, cons: List[ConstraintData]):
        """
        Add constraints to the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'add_constraints'."
        )

    def add_block(self, block: BlockData):
        """
        Add a block to the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'add_block'."
        )

    def remove_constraints(self, cons: List[ConstraintData]):
        """
        Remove constraints from the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'remove_constraints'."
        )

    def remove_block(self, block: BlockData):
        """
        Remove a block from the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'remove_block'."
        )

    def update_variables(self, variables: List[VarData]):
        """
        Update variables on the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'update_variables'."
        )

    def update_parameters(self):
        """
        Update parameters on the model.
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'update_parameters'."
        )


class LegacySolverWrapper:
    """
    Class to map the new solver interface features into the legacy solver
    interface. Necessary for backwards compatibility.
    """

    class _all_true:
        # A mockup of Bunch that returns True for all attribute lookups
        # or containment tests.
        def __getattr__(self, name):
            return True

        def __contains__(self, name):
            return True

    class _UpdatableConfigDict(ConfigDict):
        # The legacy solver interface used Option objects.  we will make
        # the ConfigDict *look* like an Option by supporting update()
        __slots__ = ()

        def update(self, value):
            return self.set_value(value)

    def __init__(self, **kwargs):
        solver_io = kwargs.pop('solver_io', None)
        if solver_io:
            raise NotImplementedError(f'Still working on this ({solver_io=})')
        # There is no reason for a user to be trying to mix both old
        # and new options. That is silly. So we will yell at them.
        _options = kwargs.pop('options', None)
        if 'solver_options' in kwargs:
            if _options is not None:
                raise ValueError(
                    "Both 'options' and 'solver_options' were requested. "
                    "Please use one or the other, not both."
                )
            _options = kwargs.pop('solver_options')
        if _options is not None:
            kwargs['solver_options'] = _options
        super().__init__(**kwargs)
        # Make the legacy 'options' attribute an alias of the new
        # config.solver_options; change its class so it behaves more
        # like an old Bunch/Options object.
        self.options = self.config.solver_options
        self.options.__class__ = LegacySolverWrapper._UpdatableConfigDict
        # We will assume the solver interface supports all capabilities
        # (unless a derived class actually set something
        if not hasattr(self, '_capabilities'):
            self._capabilities = LegacySolverWrapper._all_true()

    #
    # Support "with" statements
    #
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exit statement - enables `with` statements."""

    def __setattr__(self, attr, value):
        # 'options' and 'config' are really singleton attributes.  Map
        # any assignment to set_value()
        if attr in ('options', 'config') and attr in self.__dict__:
            getattr(self, attr).set_value(value)
        else:
            super().__setattr__(attr, value)

    def _map_config(
        self,
        tee=NOTSET,
        load_solutions=NOTSET,
        symbolic_solver_labels=NOTSET,
        timelimit=NOTSET,
        report_timing=NOTSET,
        raise_exception_on_nonoptimal_result=NOTSET,
        solver_io=NOTSET,
        suffixes=NOTSET,
        logfile=NOTSET,
        keepfiles=NOTSET,
        solnfile=NOTSET,
        options=NOTSET,
        solver_options=NOTSET,
        writer_config=NOTSET,
    ):
        """Map between legacy and new interface configuration options"""
        if 'report_timing' not in self.config:
            self.config.declare(
                'report_timing', ConfigValue(domain=bool, default=False)
            )
        if tee is not NOTSET:
            self.config.tee = tee
        if load_solutions is not NOTSET:
            self.config.load_solutions = load_solutions
        if symbolic_solver_labels is not NOTSET:
            self.config.symbolic_solver_labels = symbolic_solver_labels
        if timelimit is not NOTSET:
            self.config.time_limit = timelimit
        if report_timing is not NOTSET:
            self.config.report_timing = report_timing
        if (options is not NOTSET) and (solver_options is not NOTSET):
            # There is no reason for a user to be trying to mix both old
            # and new options. That is silly. So we will yell at them.
            # Example that would raise an error:
            # solver.solve(model, options={'foo' : 'bar'}, solver_options={'foo' : 'not_bar'})
            raise ValueError(
                "Both 'options' and 'solver_options' were requested. "
                "Please use one or the other, not both."
            )
        if options is not NOTSET:
            # This block is trying to mimic the existing logic in the legacy
            # interface that allows users to pass initialized options to
            # the solver object and override them in the solve call.
            self.config.solver_options.set_value(options)
        elif solver_options is not NOTSET:
            self.config.solver_options.set_value(solver_options)
        if writer_config is not NOTSET:
            self.config.writer_config.set_value(writer_config)
        # This is a new flag in the interface. To preserve backwards compatibility,
        # its default is set to "False"
        if raise_exception_on_nonoptimal_result is not NOTSET:
            self.config.raise_exception_on_nonoptimal_result = (
                raise_exception_on_nonoptimal_result
            )
        if solver_io is not NOTSET and solver_io is not None:
            raise NotImplementedError('Still working on this')
        if suffixes is not NOTSET and suffixes is not None:
            raise NotImplementedError('Still working on this')
        if logfile is not NOTSET and logfile is not None:
            raise NotImplementedError('Still working on this')
        if keepfiles or 'keepfiles' in self.config:
            cwd = os.getcwd()
            deprecation_warning(
                "`keepfiles` has been deprecated in the new solver interface. "
                "Use `working_dir` instead to designate a directory in which files "
                f"should be generated and saved. Setting `working_dir` to `{cwd}`.",
                version='6.7.1',
            )
            self.config.working_dir = cwd
        # I believe this currently does nothing; however, it is unclear what
        # our desired behavior is for this.
        if solnfile is not NOTSET:
            if 'filename' in self.config:
                filename = os.path.splitext(solnfile)[0]
                self.config.filename = filename

    def _map_results(self, model, results):
        """Map between legacy and new Results objects"""
        legacy_results = LegacySolverResults()
        legacy_soln = LegacySolution()
        legacy_results.solver.status = legacy_solver_status_map[
            results.termination_condition
        ]
        legacy_results.solver.termination_condition = legacy_termination_condition_map[
            results.termination_condition
        ]
        legacy_soln.status = legacy_solution_status_map(results)
        legacy_results.solver.termination_message = str(results.termination_condition)
        legacy_results.problem.number_of_constraints = float('nan')
        legacy_results.problem.number_of_variables = float('nan')
        number_of_objectives = sum(
            1
            for _ in model.component_data_objects(
                Objective, active=True, descend_into=True
            )
        )
        legacy_results.problem.number_of_objectives = number_of_objectives
        if number_of_objectives == 1:
            obj = get_objective(model)
            legacy_results.problem.sense = obj.sense

            if obj.sense == minimize:
                legacy_results.problem.lower_bound = results.objective_bound
                legacy_results.problem.upper_bound = results.incumbent_objective
            else:
                legacy_results.problem.upper_bound = results.objective_bound
                legacy_results.problem.lower_bound = results.incumbent_objective
        if (
            results.incumbent_objective is not None
            and results.objective_bound is not None
        ):
            legacy_soln.gap = abs(results.incumbent_objective - results.objective_bound)
        else:
            legacy_soln.gap = None
        return legacy_results, legacy_soln

    def _solution_handler(
        self, load_solutions, model, results, legacy_results, legacy_soln
    ):
        """Method to handle the preferred action for the solution"""
        symbol_map = legacy_soln.symbol_map = SymbolMap()
        symbol_map.default_labeler = NumericLabeler('x')
        if not hasattr(model, 'solutions'):
            # This logic gets around Issue #2130 in which
            # solutions is not an attribute on Blocks
            from pyomo.core.base.PyomoModel import ModelSolutions

            setattr(model, 'solutions', ModelSolutions(model))
        model.solutions.add_symbol_map(symbol_map)
        legacy_results._smap = symbol_map
        legacy_results._smap_id = id(symbol_map)
        delete_legacy_soln = True
        if load_solutions:
            if hasattr(model, 'dual') and model.dual.import_enabled():
                for con, val in results.solution_loader.get_duals().items():
                    model.dual[con] = val
            if hasattr(model, 'rc') and model.rc.import_enabled():
                for var, val in results.solution_loader.get_reduced_costs().items():
                    model.rc[var] = val
        elif results.incumbent_objective is not None:
            delete_legacy_soln = False
            for var, val in results.solution_loader.get_primals().items():
                legacy_soln.variable[symbol_map.getSymbol(var)] = {'Value': val}
            if hasattr(model, 'dual') and model.dual.import_enabled():
                for con, val in results.solution_loader.get_duals().items():
                    legacy_soln.constraint[symbol_map.getSymbol(con)] = {'Dual': val}
            if hasattr(model, 'rc') and model.rc.import_enabled():
                for var, val in results.solution_loader.get_reduced_costs().items():
                    legacy_soln.variable['Rc'] = val

        legacy_results.solution.insert(legacy_soln)
        # Timing info was not originally on the legacy results, but we
        # want to make it accessible to folks who are utilizing the
        # backwards compatible version.  Note that embedding the
        # ConfigDict broke pickling the legacy_results, so we will only
        # return raw nested dicts
        legacy_results.timing_info = results.timing_info.value()
        if delete_legacy_soln:
            legacy_results.solution.delete(0)
        return legacy_results

    def solve(
        self,
        model: BlockData,
        tee: bool = False,
        load_solutions: bool = True,
        logfile: Optional[str] = None,
        solnfile: Optional[str] = None,
        timelimit: Optional[float] = None,
        report_timing: bool = False,
        solver_io: Optional[str] = None,
        suffixes: Optional[Sequence] = None,
        options: Optional[Dict] = None,
        keepfiles: bool = False,
        symbolic_solver_labels: bool = False,
        # These are for forward-compatibility
        raise_exception_on_nonoptimal_result: bool = False,
        solver_options: Optional[Dict] = None,
        writer_config: Optional[Dict] = None,
    ):
        """
        Solve method: maps new solve method style to backwards compatible version.

        Returns
        -------
        legacy_results
            Legacy results object

        """
        original_config = self.config

        map_args = (
            'tee',
            'load_solutions',
            'symbolic_solver_labels',
            'timelimit',
            'report_timing',
            'raise_exception_on_nonoptimal_result',
            'solver_io',
            'suffixes',
            'logfile',
            'keepfiles',
            'solnfile',
            'options',
            'solver_options',
            'writer_config',
        )
        loc = locals()
        filtered_args = {k: loc[k] for k in map_args if loc.get(k, None) is not None}
        self._map_config(**filtered_args)

        results: Results = super().solve(model)
        legacy_results, legacy_soln = self._map_results(model, results)
        legacy_results = self._solution_handler(
            load_solutions, model, results, legacy_results, legacy_soln
        )

        if self.config.report_timing:
            print(results.timing_info.timer)

        self.config = original_config

        return legacy_results

    def available(self, exception_flag=True):
        """
        Returns a bool determining whether the requested solver is available
        on the system.
        """
        ans = super().available()
        if exception_flag and not ans:
            raise ApplicationError(
                f'Solver "{self.name}" is not available. '
                f'The returned status is: {ans}.'
            )
        return bool(ans)

    def license_is_valid(self) -> bool:
        """Test if the solver license is valid on this system.

        Note that this method is included for compatibility with the
        legacy SolverFactory interface.  Unlicensed or open source
        solvers will return True by definition.  Licensed solvers will
        return True if a valid license is found.

        Returns
        -------
        available: bool
            True if the solver license is valid. Otherwise, False.

        """
        return bool(self.available())

    def config_block(self, init=False):
        """
        Preserves config backwards compatibility; allows new solver interfaces
        to be used in the pyomo solve call
        """
        return default_config_block(self, init)[0]

    def set_options(self, options):
        """
        Method to manually set options; can be called outside of the solve method
        """
        opts = {k: v for k, v in options.value().items() if v is not None}
        if opts:
            self._map_config(**opts)

    def warm_start_capable(self):
        return False

    def default_variable_value(self):
        return None

    @classmethod
    def api_version(self):
        """
        Return the public API supported by this interface.

        Returns
        -------
        ~pyomo.common.enums.SolverAPIVersion
            A solver API enum object
        """
        return SolverAPIVersion.V1



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
from os.path import dirname, abspath, join

thisDir = dirname(abspath(__file__))

import pyomo.common.unittest as unittest

import pyomo.opt
from pyomo.common.dependencies import yaml, yaml_available, yaml_load_args
from pyomo.common.fileutils import import_file
from pyomo.core.base import Var
from pyomo.core.base.objective import minimize, maximize
from pyomo.core.base.piecewise import Bound, PWRepn
from pyomo.solvers.tests.solvers import test_solver_cases as _test_solver_cases

smoke_problems = ['convex_var', 'step_var', 'step_vararray']

nightly_problems = [
    'convex_vararray',
    'concave_vararray',
    'concave_var',
    'piecewise_var',
    'piecewise_vararray',
]

expensive_problems = [
    'piecewise_multi_vararray',
    'convex_multi_vararray1',
    'concave_multi_vararray1',
    'convex_multi_vararray2',
    'concave_multi_vararray2',
]


testing_solvers = {}
testing_solvers['gurobi', 'lp'] = False
# testing_solvers['cplex','lp'] = False
# testing_solvers['cplex','nl'] = False
# testing_solvers['ipopt','nl'] = False
# testing_solvers['cplex','python'] = False
# testing_solvers['_cplex_persistent','python'] = False
for _solver, _io in _test_solver_cases():
    if (_solver, _io) in testing_solvers and _test_solver_cases(_solver, _io).available:
        testing_solvers[_solver, _io] = True


def createMethod(pName, problem, solver, writer, kwds):
    def Method(obj):
        if not testing_solvers[solver, writer]:
            obj.skipTest("Solver %s (interface=%s) is not available" % (solver, writer))

        m = import_file(
            os.path.join(thisDir, 'problems', problem + '.py'), clear_cache=True
        )

        model = m.define_model(**kwds)

        opt = pyomo.opt.SolverFactory(solver, solver_io=writer)
        results = opt.solve(model)

        # non-recursive
        new_results = (
            (var.name, var.value)
            for var in model.component_data_objects(
                Var, active=True, descend_into=False
            )
        )
        baseline_results = getattr(obj, problem + '_results')
        for name, value in new_results:
            if abs(baseline_results[name] - value) > 0.0001:
                raise IOError(
                    "Difference in baseline solution values and "
                    "current solution values using:\n"
                    + "Solver: "
                    + solver
                    + "\n"
                    + "Writer: "
                    + writer
                    + "\n"
                    + "Variable: "
                    + name
                    + "\n"
                    + "Solution: "
                    + str(value)
                    + "\n"
                    + "Baseline: "
                    + str(baseline_results[name])
                    + "\n"
                )

    return Method


def assignProblems(cls, problem_list):
    for solver, writer in testing_solvers:
        for PROBLEM in problem_list:
            aux_list = ['', 'force_pw']
            for AUX in aux_list:
                for REPN in PWRepn:
                    for BOUND_TYPE in Bound:
                        for SENSE in [maximize, minimize]:
                            if not (
                                ((BOUND_TYPE == Bound.Lower) and (SENSE == maximize))
                                or ((BOUND_TYPE == Bound.Upper) and (SENSE == minimize))
                                or (
                                    (
                                        REPN
                                        in [
                                            PWRepn.BIGM_BIN,
                                            PWRepn.BIGM_SOS1,
                                            PWRepn.MC,
                                        ]
                                    )
                                    and ('step' in PROBLEM)
                                )
                            ):
                                kwds = {}
                                kwds['sense'] = SENSE
                                kwds['pw_repn'] = REPN
                                kwds['pw_constr_type'] = BOUND_TYPE
                                if SENSE == maximize:
                                    attrName = "test_{0}_{1}_{2}_{3}_{4}_{5}".format(
                                        PROBLEM,
                                        REPN,
                                        BOUND_TYPE,
                                        'maximize',
                                        solver,
                                        writer,
                                    )
                                elif SENSE == minimize:
                                    attrName = "test_{0}_{1}_{2}_{3}_{4}_{5}".format(
                                        PROBLEM,
                                        REPN,
                                        BOUND_TYPE,
                                        'minimize',
                                        solver,
                                        writer,
                                    )
                                if AUX != '':
                                    kwds[AUX] = True
                                    attrName += '_' + AUX
                                setattr(
                                    cls,
                                    attrName,
                                    createMethod(
                                        attrName, PROBLEM, solver, writer, kwds
                                    ),
                                )
                                if yaml_available:
                                    with open(
                                        join(
                                            thisDir,
                                            'baselines',
                                            PROBLEM + '_baseline_results.yml',
                                        ),
                                        'r',
                                    ) as f:
                                        baseline_results = yaml.load(
                                            f, **yaml_load_args
                                        )
                                        setattr(
                                            cls, PROBLEM + '_results', baseline_results
                                        )


@unittest.skipUnless(yaml_available, "PyYAML module is not available.")
class PW_Tests(unittest.TestCase):
    pass


class PiecewiseLinearTest_Smoke(PW_Tests):
    pass


assignProblems(PiecewiseLinearTest_Smoke, smoke_problems)


class PiecewiseLinearTest_Nightly(PW_Tests):
    pass


assignProblems(PiecewiseLinearTest_Nightly, nightly_problems)


@unittest.pytest.mark.expensive
class PiecewiseLinearTest_Expensive(PW_Tests):
    pass


assignProblems(PiecewiseLinearTest_Expensive, expensive_problems)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
#

import pyomo.common.unittest as unittest
from pyomo.environ import (
    ConcreteModel,
    Var,
    Param,
    Set,
    Constraint,
    Objective,
    Expression,
    Suffix,
    RangeSet,
    ExternalFunction,
    units,
    maximize,
    sin,
    cos,
    sqrt,
)
from pyomo.network import Port, Arc
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.gdp import Disjunct, Disjunction
from pyomo.core.base.units_container import pint_available, UnitsError
from pyomo.util.check_units import (
    assert_units_consistent,
    assert_units_equivalent,
    check_units_equivalent,
    identify_inconsistent_units,
)


def python_callback_function(arg1, arg2):
    return 42.0


@unittest.skipIf(not pint_available, "Testing units requires pint")
class TestUnitsChecking(unittest.TestCase):
    def _create_model_and_vars(self):
        u = units
        m = ConcreteModel()
        m.dx = Var(units=u.m, initialize=0.10188943773836046)
        m.dy = Var(units=u.m, initialize=0.0)
        m.vx = Var(units=u.m / u.s, initialize=0.7071067769802851)
        m.vy = Var(units=u.m / u.s, initialize=0.7071067769802851)
        m.t = Var(units=u.s, bounds=(1e-5, 10.0), initialize=0.0024015570927624456)
        m.theta = Var(
            bounds=(0, 0.49 * 3.14), initialize=0.7853981693583533, units=u.radians
        )
        m.a = Param(initialize=-32.2, units=u.ft / u.s**2)
        m.x_unitless = Var()
        return m

    def test_assert_units_consistent_equivalent(self):
        u = units
        m = ConcreteModel()
        m.dx = Var(units=u.m, initialize=0.10188943773836046)
        m.dy = Var(units=u.m, initialize=0.0)
        m.vx = Var(units=u.m / u.s, initialize=0.7071067769802851)
        m.vy = Var(units=u.m / u.s, initialize=0.7071067769802851)
        m.t = Var(units=u.min, bounds=(1e-5, 10.0), initialize=0.0024015570927624456)
        m.theta = Var(
            bounds=(0, 0.49 * 3.14), initialize=0.7853981693583533, units=u.radians
        )
        m.a = Param(initialize=-32.2, units=u.ft / u.s**2)
        m.x_unitless = Var()

        m.obj = Objective(expr=m.dx, sense=maximize)
        m.vx_con = Constraint(expr=m.vx == 1.0 * u.m / u.s * cos(m.theta))
        m.vy_con = Constraint(expr=m.vy == 1.0 * u.m / u.s * sin(m.theta))
        m.dx_con = Constraint(expr=m.dx == m.vx * u.convert(m.t, to_units=u.s))
        m.dy_con = Constraint(
            expr=m.dy
            == m.vy * u.convert(m.t, to_units=u.s)
            + 0.5
            * (u.convert(m.a, to_units=u.m / u.s**2))
            * (u.convert(m.t, to_units=u.s)) ** 2
        )
        m.ground = Constraint(expr=m.dy == 0)
        m.unitless_con = Constraint(expr=m.x_unitless == 5.0)

        assert_units_consistent(m)  # check model
        assert_units_consistent(m.dx)  # check var - this should never fail
        assert_units_consistent(
            m.x_unitless
        )  # check unitless var - this should never fail
        assert_units_consistent(m.vx_con)  # check constraint
        assert_units_consistent(m.unitless_con)  # check unitless constraint

        assert_units_equivalent(m.dx, m.dy)  # check var
        assert_units_equivalent(m.x_unitless, u.dimensionless)  # check unitless var
        assert_units_equivalent(m.x_unitless, None)  # check unitless var
        assert_units_equivalent(m.vx_con.body, u.m / u.s)  # check constraint
        assert_units_equivalent(
            m.unitless_con.body, u.dimensionless
        )  # check unitless constraint
        assert_units_equivalent(m.dx, m.dy)  # check var
        assert_units_equivalent(m.x_unitless, u.dimensionless)  # check unitless var
        assert_units_equivalent(m.x_unitless, None)  # check unitless var
        assert_units_equivalent(m.vx_con.body, u.m / u.s)  # check constraint

        m.broken = Constraint(expr=m.dy == 42.0 * u.kg)
        with self.assertRaises(UnitsError):
            assert_units_consistent(m)
        assert_units_consistent(m.dx)
        assert_units_consistent(m.vx_con)
        with self.assertRaises(UnitsError):
            assert_units_consistent(m.broken)

        self.assertTrue(check_units_equivalent(m.dx, m.dy))
        self.assertFalse(check_units_equivalent(m.dx, m.vx))

    def test_assert_units_consistent_on_datas(self):
        u = units
        m = ConcreteModel()
        m.S = Set(initialize=[1, 2, 3])
        m.x = Var(m.S, units=u.m)
        m.t = Var(m.S, units=u.s)
        m.v = Var(m.S, units=u.m / u.s)
        m.unitless = Var(m.S)

        @m.Constraint(m.S)
        def vel_con(m, i):
            return m.v[i] == m.x[i] / m.t[i]

        @m.Constraint(m.S)
        def unitless_con(m, i):
            return m.unitless[i] == 42.0

        @m.Constraint(m.S)
        def sqrt_con(m, i):
            return sqrt(m.v[i]) == sqrt(m.x[i] / m.t[i])

        assert_units_consistent(m)  # check model
        assert_units_consistent(m.x)  # check var
        assert_units_consistent(m.t)  # check var
        assert_units_consistent(m.v)  # check var
        assert_units_consistent(m.unitless)  # check var
        assert_units_consistent(m.vel_con)  # check constraint
        assert_units_consistent(m.unitless_con)  # check unitless constraint

        assert_units_consistent(m.x[2])  # check var data
        assert_units_consistent(m.t[2])  # check var data
        assert_units_consistent(m.v[2])  # check var data
        assert_units_consistent(m.unitless[2])  # check var
        assert_units_consistent(m.vel_con[2])  # check constraint data
        assert_units_consistent(m.unitless_con[2])  # check unitless constraint data

        assert_units_equivalent(m.x[2], m.x[1])  # check var data
        assert_units_equivalent(m.t[2], u.s)  # check var data
        assert_units_equivalent(m.v[2], u.m / u.s)  # check var data
        assert_units_equivalent(
            m.unitless[2], u.dimensionless
        )  # check var data unitless
        assert_units_equivalent(m.unitless[2], None)  # check var
        assert_units_equivalent(m.vel_con[2].body, u.m / u.s)  # check constraint data
        assert_units_equivalent(
            m.unitless_con[2].body, u.dimensionless
        )  # check unitless constraint data

        @m.Constraint(m.S)
        def broken(m, i):
            return m.x[i] == 42.0 * m.v[i]

        with self.assertRaises(UnitsError):
            assert_units_consistent(m)
        with self.assertRaises(UnitsError):
            assert_units_consistent(m.broken)
        with self.assertRaises(UnitsError):
            assert_units_consistent(m.broken[1])

        # all of these should still work
        assert_units_consistent(m.x)  # check var
        assert_units_consistent(m.t)  # check var
        assert_units_consistent(m.v)  # check var
        assert_units_consistent(m.unitless)  # check var
        assert_units_consistent(m.vel_con)  # check constraint
        assert_units_consistent(m.unitless_con)  # check unitless constraint

        assert_units_consistent(m.x[2])  # check var data
        assert_units_consistent(m.t[2])  # check var data
        assert_units_consistent(m.v[2])  # check var data
        assert_units_consistent(m.unitless[2])  # check var
        assert_units_consistent(m.vel_con[2])  # check constraint data
        assert_units_consistent(m.unitless_con[2])  # check unitless constraint data

    def test_assert_units_consistent_all_components(self):
        # test all scalar components consistent
        u = units
        m = self._create_model_and_vars()
        m.obj = Objective(expr=m.dx / m.t - m.vx)
        m.con = Constraint(expr=m.dx / m.t == m.vx)
        # vars already added
        m.exp = Expression(expr=m.dx / m.t - m.vx)
        m.suff = Suffix(direction=Suffix.LOCAL)
        # params already added
        # sets already added
        m.rs = RangeSet(5)
        m.disj1 = Disjunct()
        m.disj1.constraint = Constraint(expr=m.dx / m.t <= m.vx)
        m.disj2 = Disjunct()
        m.disj2.constraint = Constraint(expr=m.dx / m.t <= m.vx)
        m.disjn = Disjunction(expr=[m.disj1, m.disj2])
        # block tested as part of model
        m.extfn = ExternalFunction(
            python_callback_function, units=u.m / u.s, arg_units=[u.m, u.s]
        )
        m.conext = Constraint(expr=m.extfn(m.dx, m.t) - m.vx == 0)
        m.cset = ContinuousSet(bounds=(0, 1))
        m.svar = Var(m.cset, units=u.m)
        m.dvar = DerivativeVar(sVar=m.svar, units=u.m / u.s)

        def prt1_rule(m):
            return {"avar": m.dx}

        def prt2_rule(m):
            return {"avar": m.dy}

        m.prt1 = Port(rule=prt1_rule)
        m.prt2 = Port(rule=prt2_rule)

        def arcrule(m):
            return dict(source=m.prt1, destination=m.prt2)

        m.arc = Arc(rule=arcrule)

        # complementarities do not work yet
        # The expression system removes the u.m since it is multiplied by zero.
        # We need to change the units_container to allow 0 when comparing units
        # m.compl = Complementarity(expr=complements(m.dx/m.t >= m.vx, m.dx == 0*u.m))

        assert_units_consistent(m)

    def test_units_roundoff_error(self):
        # Issue 2393: this example resulted in roundoff error where the
        # computed units for var_1 were
        # (0.9999999999999986,
        # <Unit('kilogram ** 1 / kelvin / meter ** 1.66533e-16 / second ** 3')>
        # )
        m = ConcreteModel()
        m.var_1 = Var(
            initialize=400,
            units=(
                (units.J**0.4)
                * (units.kg**0.2)
                * (units.W**0.6)
                / units.K
                / (units.m**2.2)
                / (units.Pa**0.2)
                / (units.s**0.8)
            ),
        )
        m.var_1.fix()
        m.var_2 = Var(initialize=400, units=units.kg / units.s**3 / units.K)
        assert_units_equivalent(m.var_1, m.var_2)

    def test_identify_inconsistent_units(self):
        u = units
        m = ConcreteModel()
        m.S = Set(initialize=[1, 2, 3])
        m.x = Var(units=u.m)
        m.t = Var(units=u.s)

        # Scalar constraints
        m.c1 = Constraint(expr=m.x == 10 * u.m)  # Units consistent
        m.c2 = Constraint(expr=m.x == m.t)  # Units inconsistent

        # Indexed Constraint
        @m.Constraint(m.S)
        def c3(blk, i):
            if i == 1:
                return m.t == 10 * u.m  # Inconsistent units
            return m.t == 10 * u.s  # Consistent units

        # Scalar Expressions
        m.e1 = Expression(expr=m.x + 10 * u.m)  # Units consistent
        m.e2 = Expression(expr=m.x + m.t)  # Units inconsistent

        # Indexed Expression
        @m.Expression(m.S)
        def e3(blk, i):
            if i == 1:
                return m.t + 10 * u.m  # Inconsistent units
            return m.t + 10 * u.s  # Consistent units

        # Scalar Objectives
        m.o1 = Objective(expr=m.x + 10 * u.m)  # Units consistent
        m.o2 = Objective(expr=m.x + m.t)  # Units inconsistent

        # Indexed Objective
        @m.Objective(m.S)
        def o3(blk, i):
            if i == 1:
                return m.t + 10 * u.m  # Inconsistent units
            return m.t + 10 * u.s  # Consistent units

        failures = identify_inconsistent_units(m)

        assert len(failures) == 6
        assert m.c2 in failures
        assert m.c3[1] in failures
        assert m.e2 in failures
        assert m.e3[1] in failures
        assert m.o2 in failures
        assert m.o3[1] in failures


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.contrib.pynumero.interfaces.external_grey_box import (
    ExternalGreyBoxModel,
    ExternalGreyBoxBlock,
)
import scipy.sparse as spa
import numpy as np
import math

"""
This module builds a simple model for use in parameter 
estimation. The module contains both the pyomo only version
of the model and the external grey box version of the model

The models are setup for nonlinear least-squares to estimate
the value of the overall heat transfer coefficient UA

# The model is as follows:
# dT1 = Th_in - Tc_out
# dT2 = Th_out - Tc_in
# lmtd = (dT2 - dT1)/ln(dT2/dT1)
# Q = UA*lmtd
# Q = Fh * Cp_h * (Th_in - Th_out))
# Q = -Fc * Cp_c * (Tc_in - Tc_out))

and written as the following for the external model
# Th_in - Tc_out - dT1 = 0
# Th_out - Tc_in - dT2 = 0
# lmtd*ln(dT2/dT1) - (dT2 - dT1) = 0
# UA*lmtd - Q = 0
# Fh * Cp_h * (Th_in - Th_out)) - Q = 0
# -Fc * Cp_c * (Tc_in - Tc_out)) = 0

"""


def build_single_point_model_pyomo_only(m):
    # fixed parameters
    m.Cp_h = 2131  # heat cap hot
    m.Cp_c = 4178  # heat cap cold
    m.Fh = 0.1  # flow hot
    m.Fc = 0.2  # flow cold

    # model inputs
    m.Th_in = pyo.Var(initialize=100, bounds=(10, None))
    m.Th_out = pyo.Var(initialize=50, bounds=(10, None))
    m.Tc_in = pyo.Var(initialize=30, bounds=(10, None))
    m.Tc_out = pyo.Var(initialize=50, bounds=(10, None))
    m.UA = pyo.Var(initialize=100)
    m.Q = pyo.Var(initialize=10000, bounds=(0, None))
    m.lmtd = pyo.Var(initialize=20, bounds=(0, None))
    m.dt1 = pyo.Var(initialize=20, bounds=(0, None))
    m.dt2 = pyo.Var(initialize=20, bounds=(0, None))

    # model constraints
    m.dt1_con = pyo.Constraint(expr=m.dt1 == m.Th_in - m.Tc_out)
    m.dt2_con = pyo.Constraint(expr=m.dt2 == m.Th_out - m.Tc_in)
    m.lmtd_con = pyo.Constraint(expr=m.lmtd * pyo.log(m.dt2 / m.dt1) == (m.dt2 - m.dt1))

    m.ua_con = pyo.Constraint(expr=m.Q == m.UA * m.lmtd)
    m.Qh_con = pyo.Constraint(expr=m.Q == m.Fh * m.Cp_h * (m.Th_in - m.Th_out))
    m.Qc_con = pyo.Constraint(expr=m.Q == m.Fc * m.Cp_c * (m.Tc_out - m.Tc_in))


def build_single_point_model_external(m):
    ex_model = UAModelExternal()
    m.egb = ExternalGreyBoxBlock()
    m.egb.set_external_model(ex_model)


class UAModelExternal(ExternalGreyBoxModel):
    def __init__(self):
        super(UAModelExternal, self).__init__()
        self._input_names = [
            'Th_in',
            'Th_out',
            'Tc_in',
            'Tc_out',
            'UA',
            'Q',
            'lmtd',
            'dT1',
            'dT2',
        ]
        self._input_values = np.zeros(self.n_inputs(), dtype=np.float64)
        self._eq_constraint_names = [
            'dT1_con',
            'dT2_con',
            'lmtd_con',
            'QUA_con',
            'Qhot_con',
            'Qcold_con',
        ]
        self._eq_constraint_multipliers = np.zeros(
            self.n_equality_constraints(), dtype=np.float64
        )

        # parameters
        self._Cp_h = 2131
        self._Cp_c = 4178
        self._Fh = 0.1
        self._Fc = 0.2

    def n_inputs(self):
        return len(self.input_names())

    def n_equality_constraints(self):
        return len(self.equality_constraint_names())

    def n_outputs(self):
        return 0

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return self._eq_constraint_names

    def output_names(self):
        return []

    def finalize_block_construction(self, pyomo_block):
        pyomo_block.inputs['Th_in'].setlb(10)
        pyomo_block.inputs['Th_in'].set_value(100)
        pyomo_block.inputs['Th_out'].setlb(10)
        pyomo_block.inputs['Th_out'].set_value(50)

        pyomo_block.inputs['Tc_in'].setlb(10)
        pyomo_block.inputs['Tc_in'].set_value(30)
        pyomo_block.inputs['Tc_out'].setlb(10)
        pyomo_block.inputs['Tc_out'].set_value(50)

        pyomo_block.inputs['UA'].set_value(100)

        pyomo_block.inputs['Q'].setlb(0)
        pyomo_block.inputs['Q'].set_value(10000)

        pyomo_block.inputs['lmtd'].setlb(0)
        pyomo_block.inputs['lmtd'].set_value(20)

        pyomo_block.inputs['dT1'].setlb(0)
        pyomo_block.inputs['dT1'].set_value(20)
        pyomo_block.inputs['dT2'].setlb(0)
        pyomo_block.inputs['dT2'].set_value(20)

    def set_input_values(self, input_values):
        assert len(input_values) == 9
        np.copyto(self._input_values, input_values)

    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        assert len(eq_con_multiplier_values) == 6
        np.copyto(self._eq_constraint_multipliers, eq_con_multiplier_values)

    def evaluate_equality_constraints(self):
        Th_in = self._input_values[0]
        Th_out = self._input_values[1]
        Tc_in = self._input_values[2]
        Tc_out = self._input_values[3]
        UA = self._input_values[4]
        Q = self._input_values[5]
        lmtd = self._input_values[6]
        dT1 = self._input_values[7]
        dT2 = self._input_values[8]

        resid = np.zeros(self.n_equality_constraints())
        resid[0] = Th_in - Tc_out - dT1
        resid[1] = Th_out - Tc_in - dT2
        resid[2] = lmtd * math.log(dT2 / dT1) - (dT2 - dT1)
        resid[3] = UA * lmtd - Q
        resid[4] = self._Fh * self._Cp_h * (Th_in - Th_out) - Q
        resid[5] = -self._Fc * self._Cp_c * (Tc_in - Tc_out) - Q

        return resid

    def evaluate_jacobian_equality_constraints(self):
        Th_in = self._input_values[0]
        Th_out = self._input_values[1]
        Tc_in = self._input_values[2]
        Tc_out = self._input_values[3]
        UA = self._input_values[4]
        Q = self._input_values[5]
        lmtd = self._input_values[6]
        dT1 = self._input_values[7]
        dT2 = self._input_values[8]

        row = np.zeros(18, dtype=np.int64)
        col = np.zeros(18, dtype=np.int64)
        data = np.zeros(18, dtype=np.float64)

        idx = 0
        row[idx], col[idx], data[idx] = (0, 0, 1.0)
        idx += 1
        row[idx], col[idx], data[idx] = (0, 3, -1.0)
        idx += 1
        row[idx], col[idx], data[idx] = (0, 7, -1.0)
        idx += 1
        row[idx], col[idx], data[idx] = (1, 1, 1.0)
        idx += 1
        row[idx], col[idx], data[idx] = (1, 2, -1.0)
        idx += 1
        row[idx], col[idx], data[idx] = (1, 8, -1.0)
        idx += 1
        row[idx], col[idx], data[idx] = (2, 6, math.log(dT2 / dT1))
        idx += 1
        row[idx], col[idx], data[idx] = (2, 7, -lmtd / dT1 + 1)
        idx += 1
        row[idx], col[idx], data[idx] = (2, 8, lmtd / dT2 - 1)
        idx += 1
        row[idx], col[idx], data[idx] = (3, 4, lmtd)
        idx += 1
        row[idx], col[idx], data[idx] = (3, 5, -1.0)
        idx += 1
        row[idx], col[idx], data[idx] = (3, 6, UA)
        idx += 1
        row[idx], col[idx], data[idx] = (4, 0, self._Fh * self._Cp_h)
        idx += 1
        row[idx], col[idx], data[idx] = (4, 1, -self._Fh * self._Cp_h)
        idx += 1
        row[idx], col[idx], data[idx] = (4, 5, -1)
        idx += 1
        row[idx], col[idx], data[idx] = (5, 2, -self._Fc * self._Cp_c)
        idx += 1
        row[idx], col[idx], data[idx] = (5, 3, self._Fc * self._Cp_c)
        idx += 1
        row[idx], col[idx], data[idx] = (5, 5, -1.0)
        idx += 1
        assert idx == 18

        return spa.coo_matrix((data, (row, col)), shape=(6, 9))

    def evaluate_hessian_equality_constraints(self):
        Th_in = self._input_values[0]
        Th_out = self._input_values[1]
        Tc_in = self._input_values[2]
        Tc_out = self._input_values[3]
        UA = self._input_values[4]
        Q = self._input_values[5]
        lmtd = self._input_values[6]
        dT1 = self._input_values[7]
        dT2 = self._input_values[8]

        row = np.zeros(5, dtype=np.int64)
        col = np.zeros(5, dtype=np.int64)
        data = np.zeros(5, dtype=np.float64)
        lam = self._eq_constraint_multipliers

        idx = 0
        # lmtd_con
        row[idx], col[idx], data[idx] = (7, 6, lam[2] * (-1) / dT1)
        idx += 1
        row[idx], col[idx], data[idx] = (7, 7, lam[2] * lmtd / (dT1**2))
        idx += 1
        row[idx], col[idx], data[idx] = (8, 6, lam[2] * 1 / dT2)
        idx += 1
        row[idx], col[idx], data[idx] = (8, 8, lam[2] * (-lmtd) / (dT2**2))
        idx += 1

        # QUA_con
        row[idx], col[idx], data[idx] = (6, 4, lam[3] * (1))
        idx += 1
        assert idx == 5

        return spa.coo_matrix((data, (row, col)), shape=(9, 9))

    #
    # Implement the following methods to provide support for
    # Hessian computations: see documentation in class docstring
    #
    # def evaluate_hessian_equality_constraints(self):
    # def evaluate_hessian_outputs(self):
    #



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import logging

from pyomo.common import Executable
from pyomo.common.collections import Bunch
from pyomo.common.dependencies import pathlib
from pyomo.opt.base.formats import ProblemFormat, ResultsFormat
from pyomo.opt.base.solvers import SolverFactory, OptSolver
from pyomo.solvers.plugins.solvers.ASL import ASL

logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register(
    'knitroampl', doc='The Knitro solver for NLP/MINLP and their subclasses'
)
class KNITROAMPL(ASL):
    """An interface to the Knitro optimizer that uses the AMPL Solver Library."""

    def __init__(self, **kwds):
        """Constructor"""
        executable = kwds.pop('executable', None)
        validate = kwds.pop('validate', True)
        kwds["type"] = "knitroampl"
        kwds["options"] = {'solver': "knitroampl"}
        OptSolver.__init__(self, **kwds)
        self._keepfiles = False
        self._results_file = None
        self._timer = ''
        self._user_executable = None
        # broadly useful for reporting, and in cases where
        # a solver plugin may not report execution time.
        self._last_solve_time = None
        self._define_signal_handlers = None
        self._version_timeout = 2

        if executable == 'knitroampl':
            self.set_executable(name=None, validate=validate)
        elif executable is not None:
            self.set_executable(name=executable, validate=validate)
        #
        # Setup valid problem formats, and valid results for each problem format.
        # Also set the default problem and results formats.
        #
        self._valid_problem_formats = [ProblemFormat.nl]
        self._valid_result_formats = {ProblemFormat.nl: [ResultsFormat.sol]}
        self.set_problem_format(ProblemFormat.nl)
        #
        # Note: Undefined capabilities default to 'None'
        #
        self._capabilities = Bunch()
        self._capabilities.linear = True
        self._capabilities.integer = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.sos1 = False
        self._capabilities.sos2 = False

    def _default_executable(self):
        try:
            # If knitro Python package is available, use the executable it contains
            import knitro

            package_knitroampl_path = (
                pathlib.Path(knitro.__file__).resolve().parent
                / 'knitroampl'
                / 'knitroampl'
            )
            executable = Executable(str(package_knitroampl_path))
        except ModuleNotFoundError:
            # Otherwise, search usual path list
            executable = Executable('knitroampl')
        if not executable:
            logger.warning(
                "Could not locate the 'knitroampl' executable, "
                "which is required for solver %s" % self.name
            )
            self.enable = False
            return None
        return executable.path()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from pyomo.repn.standard_repn import (
    preprocess_block_constraints,
    preprocess_block_objectives,
)


def compute_standard_repn(data, model=None):
    """
    This plugin computes the standard representation for all objectives
    and constraints. All results are stored in a ComponentMap named
    "_repn" at the block level.

    We break out preprocessing of the objectives and constraints
    in order to avoid redundant and unnecessary work, specifically
    in contexts where a model is iteratively solved and modified.
    we don't have finer-grained resolution, but we could easily
    pass in a Constraint and an Objective if warranted.

    Required:
        model:      A concrete model instance.
    """
    idMap = {}
    for block in model.block_data_objects(active=True):
        preprocess_block_constraints(block, idMap=idMap)
        preprocess_block_objectives(block, idMap=idMap)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import os
import re
import sys
import time

from pyomo.common.collections import ComponentSet, ComponentMap, Bunch
from pyomo.common.dependencies import attempt_import
from pyomo.common.errors import ApplicationError
from pyomo.common.tee import capture_output
from pyomo.common.tempfiles import TempfileManager
from pyomo.core.expr.numvalue import is_fixed
from pyomo.core.expr.numvalue import value
from pyomo.core.staleflag import StaleFlagManager
from pyomo.repn import generate_standard_repn
from pyomo.solvers.plugins.solvers.direct_solver import DirectSolver
from pyomo.solvers.plugins.solvers.direct_or_persistent_solver import (
    DirectOrPersistentSolver,
)
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.opt.results.results_ import SolverResults
from pyomo.opt.results.solution import Solution, SolutionStatus
from pyomo.opt.results.solver import TerminationCondition, SolverStatus
from pyomo.opt.base import SolverFactory
from pyomo.core.base.suffix import Suffix
import pyomo.core.base.var

logger = logging.getLogger('pyomo.solvers')


class DegreeError(ValueError):
    pass


def _is_convertible(conv_type, x):
    try:
        conv_type(x)
    except ValueError:
        return False
    return True


def _print_message(xp_prob, _, msg, *args):
    if msg is not None:
        sys.stdout.write(msg + '\n')
        sys.stdout.flush()


def _finalize_xpress_import(xpress, avail):
    if not avail:
        return
    xp = xpress
    XpressDirect._version = tuple(int(k) for k in xp.getversion().split('.'))
    XpressDirect._name = "Xpress %s.%s.%s" % XpressDirect._version
    # In (pypi) versions prior to 8.13.0, the 'xp.rng' keyword was
    # 'xp.range'
    if not hasattr(xp, 'rng'):
        xp.rng = xp.range

    # Xpress 9.6 (45.1.1) renamed many of its enums, deprecating the old ones
    if XpressDirect._version < (45,):
        XpressDirect.LPStatus.UNSTARTED = xp.lp_unstarted
        XpressDirect.LPStatus.OPTIMAL = xp.lp_optimal
        XpressDirect.LPStatus.INFEAS = xp.lp_infeas
        XpressDirect.LPStatus.CUTOFF = xp.lp_cutoff
        XpressDirect.LPStatus.NONCONVEX = xp.lp_nonconvex
        XpressDirect.LPStatus.UNFINISHED = xp.lp_unfinished
        XpressDirect.LPStatus.UNBOUNDED = xp.lp_unbounded
        XpressDirect.LPStatus.UNSOLVED = xp.lp_unsolved
        XpressDirect.LPStatus.CUTOFF_IN_DUAL = xp.lp_cutoff_in_dual
        XpressDirect.MIPStatus.INFEAS = xp.mip_infeas
        XpressDirect.MIPStatus.LP_NOT_OPTIMAL = xp.mip_lp_not_optimal
        XpressDirect.MIPStatus.LP_OPTIMAL = xp.mip_lp_optimal
        XpressDirect.MIPStatus.NO_SOL_FOUND = xp.mip_no_sol_found
        XpressDirect.MIPStatus.NOT_LOADED = xp.mip_not_loaded
        XpressDirect.MIPStatus.OPTIMAL = xp.mip_optimal
        XpressDirect.MIPStatus.SOLUTION = xp.mip_solution
        XpressDirect.MIPStatus.UNBOUNDED = xp.mip_unbounded
        XpressDirect.NLPStatus.OPTIMAL = xp.nlp_globally_optimal
        XpressDirect.NLPStatus.INFEASIBLE = xp.nlp_infeasible
        XpressDirect.NLPStatus.LOCALLY_INFEASIBLE = xp.nlp_locally_infeasible
        XpressDirect.NLPStatus.LOCALLY_OPTIMAL = xp.nlp_locally_optimal
        XpressDirect.NLPStatus.SOLUTION = xp.nlp_solution
        XpressDirect.NLPStatus.UNBOUNDED = xp.nlp_unbounded
        XpressDirect.NLPStatus.UNFINISHED = xp.nlp_unfinished
        XpressDirect.NLPStatus.UNSTARTED = xp.nlp_unstarted
    else:
        XpressDirect.LPStatus = xp.LPStatus
        XpressDirect.MIPStatus = xp.MIPStatus
        XpressDirect.NLPStatus.OPTIMAL = xp.constants.NLPSTATUS_OPTIMAL
        XpressDirect.NLPStatus.INFEASIBLE = xp.constants.NLPSTATUS_INFEASIBLE
        XpressDirect.NLPStatus.LOCALLY_OPTIMAL = xp.constants.NLPSTATUS_LOCALLY_OPTIMAL
        XpressDirect.NLPStatus.SOLUTION = xp.constants.NLPSTATUS_SOLUTION
        XpressDirect.NLPStatus.UNBOUNDED = xp.constants.NLPSTATUS_UNBOUNDED
        XpressDirect.NLPStatus.UNFINISHED = xp.constants.NLPSTATUS_UNFINISHED
        XpressDirect.NLPStatus.UNSTARTED = xp.constants.NLPSTATUS_UNSTARTED
        XpressDirect.NLPStatus.LOCALLY_INFEASIBLE = (
            xp.constants.NLPSTATUS_LOCALLY_INFEASIBLE
        )

    # Xpress 9.5 (44.1.1) changed the Python API fairly significantly.
    # We will map between the two APIs based on the version.
    if XpressDirect._version < (44,):

        def _addConstraint(
            self,
            prob,
            constraint=None,
            body=None,
            lb=None,
            ub=None,
            type=None,
            rhs=None,
            name='',
        ):
            # It's unclear what the acceptable "default" values are for
            # lb, ub, etc. (putting in the values from the documentation
            # generates errors).  We will instead use None and filter
            # out any non-None fields.
            args = {'sense': type, 'name': name}
            for field in ('constraint', 'body', 'lb', 'ub', 'rhs'):
                if locals()[field] is not None:
                    args[field] = locals()[field]
            con = xp.constraint(**args)
            prob.addConstraint(con)
            return con

        def _addVariable(self, prob, name, lb, ub, vartype):
            var = xp.var(name=name, lb=lb, ub=ub, vartype=vartype)
            prob.addVariable(var)
            return var

        def _addSOS(self, prob, indices, weights, type, name):
            con = xp.sos(indices, weights, type, name)
            prob.addSOS(con)
            return con

        XpressDirect._addConstraint = _addConstraint
        XpressDirect._addVariable = _addVariable
        XpressDirect._addSOS = _addSOS
        XpressDirect._getSlacks = lambda self, prob, con: prob.getSlack(con)
        XpressDirect._getDuals = lambda self, prob, con: prob.getDual(con)
        XpressDirect._getRedCosts = lambda self, prob, con: prob.getRCost(con)
    else:
        # Note that rhsrange (the last argument) was not added until
        # 9.5.  We will not include it here in the compatibility
        # wrapper.
        def _addConstraint(
            self,
            prob,
            constraint=None,
            body=None,
            lb=None,
            ub=None,
            type=None,
            rhs=None,
            name='',
        ):
            con = xp.constraint(
                constraint=constraint,
                body=body,
                lb=lb,
                ub=ub,
                type=type,
                rhs=rhs,
                name=name,
            )
            prob.addConstraint(con)
            return con

        XpressDirect._addConstraint = _addConstraint
        XpressDirect._addVariable = (
            lambda self, prob, name, lb, ub, vartype: prob.addVariable(
                name=name, lb=lb, ub=ub, vartype=vartype
            )
        )
        XpressDirect._addSOS = (
            lambda self, prob, indices, weights, type, name: prob.addSOS(
                indices, weights, type, name
            )
        )
        XpressDirect._getSlacks = lambda self, prob, con: prob.getSlacks(con)
        XpressDirect._getDuals = lambda self, prob, con: prob.getDuals(con)
        XpressDirect._getRedCosts = lambda self, prob, con: prob.getRedCosts(con)

        # Note that as of 9.5, xp.var raises an exception when
        # compared using '==' after it has been removed from the model.
        # This can foul up ComponentMaps in the persistent interface,
        # so we will hard-code the `var` as not being hashable (so the
        # ComponentMap will use the id() as the key)
        ComponentMap.hasher.hashable(xp.var, False)

    # Xpress 9.8 (46) adopted camelcase function naming, deprecating the old names
    if XpressDirect._version < (46,):
        XpressDirect._setLogFile = lambda self, prob, *args, **kwargs: prob.setlogfile(
            *args, **kwargs
        )
        XpressDirect._lpOptimize = lambda self, prob, *args, **kwargs: prob.lpoptimize(
            *args, **kwargs
        )
        XpressDirect._mipOptimize = (
            lambda self, prob, *args, **kwargs: prob.mipoptimize(*args, **kwargs)
        )
        XpressDirect._nlpOptimize = (
            lambda self, prob, *args, **kwargs: prob.nlpoptimize(*args, **kwargs)
        )
        XpressDirect._postSolve = lambda self, prob, *args, **kwargs: prob.postsolve(
            *args, **kwargs
        )
        XpressDirect._chgBounds = lambda sef, prob, *args, **kwargs: prob.chgbounds(
            *args, **kwargs
        )
        XpressDirect._addMipSol = lambda self, prob, *args, **kwargs: prob.addmipsol(
            *args, **kwargs
        )
        XpressDirect._addCols = lambda self, prob, objx, mstart, mrwind, dmatval, bdl, bdu, names, types: prob.addcols(
            objx, mstart, mrwind, dmatval, bdl, bdu, names, types
        )
        XpressDirect._chgColType = lambda self, prob, *args, **kwargs: prob.chgcoltype(
            *args, *kwargs
        )
        XpressDirect._getIndex = (
            lambda self, prob, *args, **kwargs: prob.getIndexFromName(*args, **kwargs)
        )
        XpressDirect._getObjIndex = lambda self, prob, obj: prob.getIndex(obj)

        def _getLB(self, prob, *args, **kwargs):
            lb = []
            prob.getlb(lb, *args, *kwargs)
            return lb

        def _getUB(self, prob, *args, **kwargs):
            ub = []
            prob.getub(ub, *args, *kwargs)
            return ub

        XpressDirect._getLB = _getLB
        XpressDirect._getUB = _getUB

    else:
        XpressDirect._setLogFile = lambda self, prob, *args, **kwargs: prob.setLogFile(
            *args, **kwargs
        )
        XpressDirect._lpOptimize = lambda self, prob, *args, **kwargs: prob.lpOptimize(
            *args, **kwargs
        )
        XpressDirect._mipOptimize = (
            lambda self, prob, *args, **kwargs: prob.mipOptimize(*args, **kwargs)
        )
        XpressDirect._nlpOptimize = (
            lambda self, prob, *args, **kwargs: prob.nlpOptimize(*args, **kwargs)
        )
        XpressDirect._postSolve = lambda self, prob, *args, **kwargs: prob.postSolve(
            *args, **kwargs
        )
        XpressDirect._chgBounds = lambda sef, prob, *args, **kwargs: prob.chgBounds(
            *args, **kwargs
        )
        XpressDirect._addMipSol = lambda self, prob, *args, **kwargs: prob.addMipSol(
            *args, **kwargs
        )
        XpressDirect._chgColType = lambda self, prob, *args, **kwargs: prob.chgColType(
            *args, *kwargs
        )
        XpressDirect._getIndex = lambda self, prob, *args, **kwargs: prob.getIndex(
            *args, **kwargs
        )
        XpressDirect._getObjIndex = lambda self, prob, obj: obj.index
        XpressDirect._getLB = lambda self, prob, *args, **kwargs: prob.getLB(
            *args, **kwargs
        )
        XpressDirect._getUB = lambda self, prob, *args, **kwargs: prob.getUB(
            *args, **kwargs
        )

        def _addCols(self, prob, objx, mstart, mrwind, dmatval, bdl, bdu, names, types):
            first_col_ind = prob.attributes.cols
            prob.addCols(objx, mstart, mrwind, dmatval, bdl, bdu)
            last_col_ind = prob.attributes.cols - 1
            if names is not None:
                prob.addNames(xp.Namespaces.COLUMN, names, first_col_ind, last_col_ind)
            if types is not None:
                col_indices = list(range(first_col_ind, last_col_ind + 1))
                prob.chgColType(col_indices, types)

        XpressDirect._addCols = _addCols


class _xpress_importer_class:
    # We want to be able to *update* the message that the deferred
    # import generates using the stdout recorded during the actual
    # import.  As strings are immutable in Python, we will give this
    # *class* as the error message, so that we can update the embedded
    # string later.
    def __init__(self):
        self.import_message = ""

    def __str__(self):
        return str(self.import_message)

    def __call__(self):
        _cwd = os.getcwd()
        try:
            with capture_output() as OUT:
                import xpress
        finally:
            # In some versions of XPRESS (notably 8.9.0), `import
            # xpress` temporarily changes the CWD.  If the import fails
            # (e.g., due to an expired license), the CWD is not always
            # restored.  This block ensures that the CWD is preserved.
            os.chdir(_cwd)
            self.import_message += OUT.getvalue()
        return xpress


@SolverFactory.register('xpress_direct', doc='Direct python interface to XPRESS')
class XpressDirect(DirectSolver):
    _name = None
    _version = None
    XpressException = RuntimeError

    class LPStatus:
        """LP Status constants compatible across Xpress versions."""

        pass

    class MIPStatus:
        """MIP Status constants compatible across Xpress versions."""

        pass

    class NLPStatus:
        """NLP Status constants compatible across Xpress versions."""

        pass

    def __init__(self, **kwds):
        if 'type' not in kwds:
            kwds['type'] = 'xpress_direct'
        super(XpressDirect, self).__init__(**kwds)
        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._solver_var_to_pyomo_var_map = ComponentMap()
        self._pyomo_con_to_solver_con_map = dict()
        self._solver_con_to_pyomo_con_map = ComponentMap()

        self._range_constraints = set()

        self._python_api_exists = xpress_available

        # TODO: this isn't a limit of XPRESS, which implements an SLP
        #       method for NLPs. But it is a limit of *this* interface
        self._max_obj_degree = 2
        self._max_constraint_degree = 2

        # There does not seem to be an easy way to get the
        # wallclock time out of xpress, so we will measure it
        # ourselves
        self._opt_time = None

        # Note: Undefined capabilities default to None
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

        # remove the instance-level definition of the xpress version:
        # because the version comes from an imported module, only one
        # version of xpress is supported (and stored as a class attribute)
        del self._version

        # xpress will apply the warmstart itself instead of
        # DirectOrPersistentSolver._presolve
        self._apply_warmstart = False

    def available(self, exception_flag=True):
        """True if the solver is available."""

        if not xpress_available:
            if exception_flag:
                xpress.log_import_warning(logger=__name__)
                raise ApplicationError(
                    "No Python bindings available for %s solver plugin" % (type(self),)
                )
            return False

        # Check that there is a valid license
        try:
            xpress.init()
            return True
        except:
            if exception_flag:
                raise
            return False
        finally:
            xpress.free()

    def _presolve(self, *args, **kwds):
        # we'll apply the warmstart in _solve_model so the
        # message "User solution (_) stored" is wrote to the
        # correct place, i.e., the console or the log or both
        self._apply_warmstart = kwds.pop("warmstart", False)
        return super()._presolve(*args, **kwds)

    def _apply_solver(self):
        StaleFlagManager.mark_all_as_stale()

        self._setLogFile(self._solver_model, self._log_file)
        if self._keepfiles:
            print("Solver log file: " + self._log_file)

        # set xpress options
        # if the user specifies a 'mipgap', set it, and
        # set xpress's related options to 0.
        if self.options.mipgap is not None:
            self._solver_model.setControl('miprelstop', float(self.options.mipgap))
            self._solver_model.setControl('miprelcutoff', 0.0)
            self._solver_model.setControl('mipaddcutoff', 0.0)
        # xpress is picky about the type which is passed
        # into a control. So we will infer and cast
        # get the xpress valid controls
        xp_controls = xpress.controls
        for key, option in self.options.items():
            if key == 'mipgap':  # handled above
                continue
            try:
                self._solver_model.setControl(key, option)
            except xpress.ModelError:
                # take another try, converting to its type
                # we'll wrap this in a function to raise the
                # xpress error
                contr_type = type(getattr(xp_controls, key))
                if not _is_convertible(contr_type, option):
                    raise
                self._solver_model.setControl(key, contr_type(option))

        start_time = time.time()
        if self._tee:
            self._solve_model()
        else:
            # In xpress versions greater than or equal 36,
            # it seems difficult to completely suppress console
            # output without disabling logging altogether.
            # As a work around, we capture all screen output
            # when tee is False.
            with capture_output() as OUT:
                self._solve_model()
        self._opt_time = time.time() - start_time

        self._setLogFile(self._solver_model, '')

        # FIXME: can we get a return code indicating if XPRESS had a
        # significant failure?
        return Bunch(rc=None, log=None)

    def _get_lb(self, var):
        """Return the upper bound associated to the pyomo variable object"""
        xp_var = self._pyomo_var_to_solver_var_map[var]
        var_idx = self._getObjIndex(self._solver_model, xp_var)
        return self._getLB(self._solver_model, var_idx, var_idx)[0]

    def _get_ub(self, var):
        """Return the upper bound associated to the pyomo variable object"""
        xp_var = self._pyomo_var_to_solver_var_map[var]
        var_idx = self._getObjIndex(self._solver_model, xp_var)
        return self._getUB(self._solver_model, var_idx, var_idx)[0]

    def _get_mip_results(self, results, soln):
        """Sets up `results` and `soln` and returns whether there is a solution
        to query.
        Returns `True` if a feasible solution is available, `False` otherwise.
        """
        xprob = self._solver_model
        xp = xpress
        xprob_attrs = xprob.attributes
        status = xprob_attrs.mipstatus
        mip_sols = xprob_attrs.mipsols
        if status == XpressDirect.MIPStatus.NOT_LOADED:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_message = (
                "Model is not loaded; no solution information is available."
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.unknown
            # no MIP solution, first LP did not solve, second LP did,
            # third search started but incomplete
        elif (
            status == XpressDirect.MIPStatus.LP_NOT_OPTIMAL
            or status == XpressDirect.MIPStatus.LP_OPTIMAL
            or status == XpressDirect.MIPStatus.NO_SOL_FOUND
        ):
            results.solver.status = SolverStatus.aborted
            results.solver.termination_message = (
                "Model is loaded, but no solution information is available."
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.unknown
        elif status == XpressDirect.MIPStatus.SOLUTION:  # some solution available
            results.solver.status = SolverStatus.warning
            results.solver.termination_message = (
                "Unable to satisfy optimality tolerances; a sub-optimal "
                "solution is available."
            )
            results.solver.termination_condition = TerminationCondition.other
            soln.status = SolutionStatus.feasible
        elif status == XpressDirect.MIPStatus.INFEAS:  # MIP proven infeasible
            results.solver.status = SolverStatus.warning
            results.solver.termination_message = "Model was proven to be infeasible"
            results.solver.termination_condition = TerminationCondition.infeasible
            soln.status = SolutionStatus.infeasible
        elif status == XpressDirect.MIPStatus.OPTIMAL:  # optimal
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Model was solved to optimality (subject to tolerances), "
                "and an optimal solution is available."
            )
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif status == XpressDirect.MIPStatus.UNBOUNDED and mip_sols > 0:
            results.solver.status = SolverStatus.warning
            results.solver.termination_message = (
                "LP relaxation was proven to be unbounded, "
                "but a solution is available."
            )
            results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        elif status == XpressDirect.MIPStatus.UNBOUNDED and mip_sols <= 0:
            results.solver.status = SolverStatus.warning
            results.solver.termination_message = (
                "LP relaxation was proven to be unbounded."
            )
            results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        else:
            results.solver.status = SolverStatus.error
            results.solver.termination_message = (
                "Unhandled Xpress solve status (" + str(status) + ")"
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error

        results.problem.upper_bound = None
        results.problem.lower_bound = None
        if xprob_attrs.objsense == 1.0:  # minimizing MIP
            try:
                results.problem.upper_bound = xprob_attrs.mipbestobjval
            except (xpress.ModelError, AttributeError):
                pass
            try:
                results.problem.lower_bound = xprob_attrs.bestbound
            except (xpress.ModelError, AttributeError):
                pass
        elif xprob_attrs.objsense == -1.0:  # maximizing MIP
            try:
                results.problem.upper_bound = xprob_attrs.bestbound
            except (xpress.ModelError, AttributeError):
                pass
            try:
                results.problem.lower_bound = xprob_attrs.mipbestobjval
            except (xpress.ModelError, AttributeError):
                pass

        return mip_sols > 0

    def _get_lp_results(self, results, soln):
        """Sets up `results` and `soln` and returns whether there is a solution
        to query.
        Returns `True` if a feasible solution is available, `False` otherwise.
        """
        xprob = self._solver_model
        xp = xpress
        xprob_attrs = xprob.attributes
        status = xprob_attrs.lpstatus
        if status == XpressDirect.LPStatus.UNSTARTED:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_message = (
                "Model is not loaded; no solution information is available."
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.unknown
        elif status == XpressDirect.LPStatus.OPTIMAL:
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Model was solved to optimality (subject to tolerances), "
                "and an optimal solution is available."
            )
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif status == XpressDirect.LPStatus.INFEAS:
            results.solver.status = SolverStatus.warning
            results.solver.termination_message = "Model was proven to be infeasible"
            results.solver.termination_condition = TerminationCondition.infeasible
            soln.status = SolutionStatus.infeasible
        elif status == XpressDirect.LPStatus.CUTOFF:
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Optimal objective for model was proven to be worse than the "
                "cutoff value specified; a solution is available."
            )
            results.solver.termination_condition = TerminationCondition.minFunctionValue
            soln.status = SolutionStatus.optimal
        elif status == XpressDirect.LPStatus.UNFINISHED:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_message = (
                "Optimization was terminated by the user."
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error
        elif status == XpressDirect.LPStatus.UNBOUNDED:
            results.solver.status = SolverStatus.warning
            results.solver.termination_message = "Model was proven to be unbounded."
            results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        elif status == XpressDirect.LPStatus.CUTOFF_IN_DUAL:
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Xpress reported the LP was cutoff in the dual."
            )
            results.solver.termination_condition = TerminationCondition.minFunctionValue
            soln.status = SolutionStatus.optimal
        elif status == XpressDirect.LPStatus.UNSOLVED:
            results.solver.status = SolverStatus.error
            results.solver.termination_message = (
                "Optimization was terminated due to unrecoverable numerical "
                "difficulties."
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error
        elif status == XpressDirect.LPStatus.NONCONVEX:
            results.solver.status = SolverStatus.error
            results.solver.termination_message = (
                "Optimization was terminated because nonconvex quadratic data "
                "were found."
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error
        else:
            results.solver.status = SolverStatus.error
            results.solver.termination_message = (
                "Unhandled Xpress solve status (" + str(status) + ")"
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error

        results.problem.upper_bound = None
        results.problem.lower_bound = None
        try:
            results.problem.upper_bound = xprob_attrs.lpobjval
            results.problem.lower_bound = xprob_attrs.lpobjval
        except (xpress.ModelError, AttributeError):
            pass

        # Not all solution information will be available in all cases, it is
        # up to the caller/user to check the actual status and figure which
        # of x, slack, duals, reduced costs are valid.
        return xprob_attrs.lpstatus in [
            XpressDirect.LPStatus.OPTIMAL,
            XpressDirect.LPStatus.CUTOFF,
            XpressDirect.LPStatus.CUTOFF_IN_DUAL,
        ]

    def _get_nlp_results(self, results, soln):
        """Sets up `results` and `soln` and returns whether there is a solution
        to query.
        Returns `True` if a feasible solution is available, `False` otherwise.
        """
        xprob = self._solver_model
        xp = xpress
        xprob_attrs = xprob.attributes
        solver = xprob_attrs.xslp_solverselected
        if solver == 2:
            # Under the hood we used the Xpress optimizer, i.e., the problem
            # was convex
            if (xprob_attrs.originalmipents > 0) or (xprob_attrs.originalsets > 0):
                return self._get_mip_results(results, soln)
            elif xprob_attrs.lpstatus and not xprob_attrs.xslp_nlpstatus:
                # If there is no NLP solver status, process the result
                # using the LP results processor.
                return self._get_lp_results(results, soln)

        # The problem was non-linear
        status = xprob_attrs.xslp_nlpstatus
        solstatus = xprob_attrs.xslp_solstatus
        have_soln = False
        optimal = False  # *globally* optimal?
        if status == XpressDirect.NLPStatus.UNSTARTED:
            results.solver.status = SolverStatus.unknown
            results.solver.termination_message = (
                "Non-convex model solve was not started"
            )
            results.solver.termination_condition = TerminationCondition.unknown
            soln.status = SolutionStatus.unknown
        elif status == XpressDirect.NLPStatus.LOCALLY_OPTIMAL:
            # This is either XpressDirect.NLPStatus.LOCALLY_OPTIMAL or XpressDirect.NLPStatus.SOLUTION
            # we must look at the solstatus to figure out which
            if solstatus in [2, 3]:
                results.solver.status = SolverStatus.ok
                results.solver.termination_message = (
                    "Non-convex model was solved to local optimality"
                )
                results.solver.termination_condition = (
                    TerminationCondition.locallyOptimal
                )
                soln.status = SolutionStatus.locallyOptimal
            else:
                results.solver.status = SolverStatus.ok
                results.solver.termination_message = (
                    "Feasible solution found for non-convex model"
                )
                results.solver.termination_condition = TerminationCondition.feasible
                soln.status = SolutionStatus.feasible
            have_soln = True
        elif status == XpressDirect.NLPStatus.OPTIMAL:
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Non-convex model was solved to global optimality"
            )
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
            have_soln = True
            optimal = True
        elif status == XpressDirect.NLPStatus.LOCALLY_INFEASIBLE:
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Non-convex model was proven to be locally infeasible"
            )
            results.solver.termination_condition = TerminationCondition.noSolution
            soln.status = SolutionStatus.unknown
        elif status == XpressDirect.NLPStatus.INFEASIBLE:
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Non-convex model was proven to be infeasible"
            )
            results.solver.termination_condition = TerminationCondition.infeasible
            soln.status = SolutionStatus.infeasible
        elif status == XpressDirect.NLPStatus.UNBOUNDED:  # locally unbounded!
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = "Non-convex model is locally unbounded"
            results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        elif status == XpressDirect.NLPStatus.UNFINISHED:
            results.solver.status = SolverStatus.ok
            results.solver.termination_message = (
                "Non-convex solve not finished (numerical issues?)"
            )
            results.solver.termination_condition = TerminationCondition.unknown
            soln.status = SolutionStatus.unknown
            have_soln = True
        else:
            results.solver.status = SolverStatus.error
            results.solver.termination_message = "Error for non-convex model: " + str(
                status
            )
            results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error

        results.problem.upper_bound = None
        results.problem.lower_bound = None
        try:
            if xprob_attrs.objsense > 0.0 or optimal:  # minimizing
                results.problem.upper_bound = xprob_attrs.xslp_objval
            if xprob_attrs.objsense < 0.0 or optimal:  # maximizing
                results.problem.lower_bound = xprob_attrs.xslp_objval
        except (xpress.ModelError, AttributeError):
            pass

        return have_soln

    def _solve_model(self):
        if self._apply_warmstart:
            self._warm_start()

        xprob = self._solver_model
        is_mip = (xprob.attributes.mipents > 0) or (xprob.attributes.sets > 0)

        # Check for quadratic objective or quadratic constraints. If there are
        # any then we call nlpoptimize since that can handle non-convex
        # quadratics as well. In case of convex quadratics it will call
        # mipoptimize under the hood.
        if (xprob.attributes.qelems > 0) or (xprob.attributes.qcelems > 0):
            self._nlpOptimize(xprob, "g" if is_mip else "")
            self._get_results = self._get_nlp_results
        elif is_mip:
            self._mipOptimize(xprob)
            self._get_results = self._get_mip_results
        else:
            self._lpOptimize(xprob)
            self._get_results = self._get_lp_results

        self._postSolve(xprob)

    def _get_expr_from_pyomo_repn(self, repn, max_degree=2):
        referenced_vars = ComponentSet()

        degree = repn.polynomial_degree()
        if (degree is None) or (degree > max_degree):
            raise DegreeError(
                'XpressDirect does not support expressions of degree {0}.'.format(
                    degree
                )
            )

        # NOTE: xpress's python interface only allows for expressions
        #       with native numeric types. Others, like numpy.float64,
        #       will cause an exception when constructing expressions
        if len(repn.linear_vars) > 0:
            referenced_vars.update(repn.linear_vars)
            new_expr = xpress.Sum(
                float(coef) * self._pyomo_var_to_solver_var_map[var]
                for coef, var in zip(repn.linear_coefs, repn.linear_vars)
            )
        else:
            new_expr = 0.0

        for coef, (x, y) in zip(repn.quadratic_coefs, repn.quadratic_vars):
            new_expr += (
                float(coef)
                * self._pyomo_var_to_solver_var_map[x]
                * self._pyomo_var_to_solver_var_map[y]
            )
            referenced_vars.add(x)
            referenced_vars.add(y)

        new_expr += repn.constant

        return new_expr, referenced_vars

    def _get_expr_from_pyomo_expr(self, expr, max_degree=2):
        if max_degree == 2:
            repn = generate_standard_repn(expr, quadratic=True)
        else:
            repn = generate_standard_repn(expr, quadratic=False)

        try:
            xpress_expr, referenced_vars = self._get_expr_from_pyomo_repn(
                repn, max_degree
            )
        except DegreeError as e:
            msg = e.args[0]
            msg += '\nexpr: {0}'.format(expr)
            raise DegreeError(msg)

        return xpress_expr, referenced_vars

    def _xpress_lb_ub_from_var(self, var):
        if var.is_fixed():
            val = var.value
            return val, val
        if var.has_lb():
            lb = value(var.lb)
        else:
            lb = -xpress.infinity
        if var.has_ub():
            ub = value(var.ub)
        else:
            ub = xpress.infinity
        return lb, ub

    def _add_var(self, var):
        varname = self._symbol_map.getSymbol(var, self._labeler)
        vartype = self._xpress_vartype_from_var(var)
        lb, ub = self._xpress_lb_ub_from_var(var)

        xpress_var = self._addVariable(
            self._solver_model, name=varname, lb=lb, ub=ub, vartype=vartype
        )

        ## bounds on binary variables don't seem to be set correctly
        ## by the method above
        if vartype == xpress.binary:
            if lb == ub:
                self._chgBounds(self._solver_model, [xpress_var], ['B'], [lb])
            else:
                self._chgBounds(
                    self._solver_model, [xpress_var, xpress_var], ['L', 'U'], [lb, ub]
                )

        self._pyomo_var_to_solver_var_map[var] = xpress_var
        self._solver_var_to_pyomo_var_map[xpress_var] = var
        self._referenced_variables[var] = 0

    def _set_instance(self, model, kwds={}):
        self._range_constraints = set()
        DirectOrPersistentSolver._set_instance(self, model, kwds)
        self._pyomo_con_to_solver_con_map = dict()
        self._solver_con_to_pyomo_con_map = ComponentMap()
        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._solver_var_to_pyomo_var_map = ComponentMap()
        try:
            if model.name is not None:
                self._solver_model = xpress.problem(name=model.name)
            else:
                self._solver_model = xpress.problem()
        except Exception:
            e = sys.exc_info()[1]
            msg = (
                "Unable to create Xpress model. "
                "Have you installed the Python "
                "bindings for Xpress?\n\n\t" + "Error message: {0}".format(e)
            )
            raise Exception(msg)
        self._add_block(model)

    def _add_block(self, block):
        DirectOrPersistentSolver._add_block(self, block)

    def _add_constraint(self, con):
        if not con.active:
            return None

        if self._skip_trivial_constraints and is_fixed(con.body):
            return None

        conname = self._symbol_map.getSymbol(con, self._labeler)

        if con._linear_canonical_form:
            xpress_expr, referenced_vars = self._get_expr_from_pyomo_repn(
                con.canonical_form(), self._max_constraint_degree
            )
        else:
            xpress_expr, referenced_vars = self._get_expr_from_pyomo_expr(
                con.body, self._max_constraint_degree
            )

        if con.has_lb():
            if not is_fixed(con.lower):
                raise ValueError(
                    "Lower bound of constraint {0} is not constant.".format(con)
                )
        if con.has_ub():
            if not is_fixed(con.upper):
                raise ValueError(
                    "Upper bound of constraint {0} is not constant.".format(con)
                )

        if con.equality:
            xpress_con = self._addConstraint(
                self._solver_model,
                body=xpress_expr,
                type=xpress.eq,
                rhs=value(con.lower),
                name=conname,
            )
        elif con.has_lb() and con.has_ub():
            xpress_con = self._addConstraint(
                self._solver_model,
                body=xpress_expr,
                type=xpress.rng,
                lb=value(con.lower),
                ub=value(con.upper),
                name=conname,
            )
            self._range_constraints.add(xpress_con)
        elif con.has_lb():
            xpress_con = self._addConstraint(
                self._solver_model,
                body=xpress_expr,
                type=xpress.geq,
                rhs=value(con.lower),
                name=conname,
            )
        elif con.has_ub():
            xpress_con = self._addConstraint(
                self._solver_model,
                body=xpress_expr,
                type=xpress.leq,
                rhs=value(con.upper),
                name=conname,
            )
        else:
            raise ValueError(
                "Constraint does not have a lower "
                "or an upper bound: {0} \n".format(con)
            )

        for var in referenced_vars:
            self._referenced_variables[var] += 1
        self._vars_referenced_by_con[con] = referenced_vars
        self._pyomo_con_to_solver_con_map[con] = xpress_con
        self._solver_con_to_pyomo_con_map[xpress_con] = con

    def _add_sos_constraint(self, con):
        if not con.active:
            return None

        conname = self._symbol_map.getSymbol(con, self._labeler)
        level = con.level
        if level not in [1, 2]:
            raise ValueError(
                "Solver does not support SOS level {0} constraints".format(level)
            )

        xpress_vars = []
        weights = []

        self._vars_referenced_by_con[con] = ComponentSet()

        if hasattr(con, 'get_items'):
            # aml sos constraint
            sos_items = list(con.get_items())
        else:
            # kernel sos constraint
            sos_items = list(con.items())

        for v, w in sos_items:
            self._vars_referenced_by_con[con].add(v)
            xpress_vars.append(self._pyomo_var_to_solver_var_map[v])
            self._referenced_variables[v] += 1
            weights.append(w)

        xpress_con = self._addSOS(
            self._solver_model, xpress_vars, weights, level, conname
        )
        self._pyomo_con_to_solver_con_map[con] = xpress_con
        self._solver_con_to_pyomo_con_map[xpress_con] = con

    def _xpress_vartype_from_var(self, var):
        """This function takes a pyomo variable and returns the appropriate
        xpress variable type

        :param var: pyomo.core.base.var.Var
        :return: xpress.continuous or xpress.binary or xpress.integer

        """
        if var.is_binary():
            vartype = xpress.binary
        elif var.is_integer():
            vartype = xpress.integer
        elif var.is_continuous():
            vartype = xpress.continuous
        else:
            raise ValueError(
                'Variable domain type is not recognized for {0}'.format(var.domain)
            )
        return vartype

    def _set_objective(self, obj):
        if self._objective is not None:
            for var in self._vars_referenced_by_obj:
                self._referenced_variables[var] -= 1
            self._vars_referenced_by_obj = ComponentSet()
            self._objective = None

        if obj.active is False:
            raise ValueError('Cannot add inactive objective to solver.')

        if obj.sense == minimize:
            sense = xpress.minimize
        elif obj.sense == maximize:
            sense = xpress.maximize
        else:
            raise ValueError('Objective sense is not recognized: {0}'.format(obj.sense))

        xpress_expr, referenced_vars = self._get_expr_from_pyomo_expr(
            obj.expr, self._max_obj_degree
        )

        for var in referenced_vars:
            self._referenced_variables[var] += 1

        self._solver_model.setObjective(xpress_expr, sense=sense)
        self._objective = obj
        self._vars_referenced_by_obj = referenced_vars

    def _postsolve(self):
        # the only suffixes that we extract from XPRESS are
        # constraint duals, constraint slacks, and variable
        # reduced-costs. scan through the solver suffix list
        # and throw an exception if the user has specified
        # any others.
        extract_duals = False
        extract_slacks = False
        extract_reduced_costs = False
        for suffix in self._suffixes:
            flag = False
            if re.match(suffix, "dual"):
                extract_duals = True
                flag = True
            if re.match(suffix, "slack"):
                extract_slacks = True
                flag = True
            if re.match(suffix, "rc"):
                extract_reduced_costs = True
                flag = True
            if not flag:
                raise RuntimeError(
                    "***The xpress_direct solver plugin cannot extract solution suffix="
                    + suffix
                )

        xprob = self._solver_model
        xp = xpress
        xprob_attrs = xprob.attributes

        ## XPRESS's status codes depend on this
        ## (number of integer vars > 0) or (number of special order sets > 0)
        is_mip = (xprob_attrs.mipents > 0) or (xprob_attrs.sets > 0)

        if is_mip:
            if extract_reduced_costs:
                logger.warning("Cannot get reduced costs for MIP.")
            if extract_duals:
                logger.warning("Cannot get duals for MIP.")
            extract_reduced_costs = False
            extract_duals = False

        self.results = SolverResults()
        soln = Solution()

        self.results.solver.name = XpressDirect._name
        self.results.solver.wallclock_time = self._opt_time

        if not hasattr(self, '_get_results'):
            raise RuntimeError(
                'Model was solved but `_get_results` property is not set'
            )
        have_soln = self._get_results(self.results, soln)
        self.results.problem.name = xprob_attrs.matrixname

        if xprob_attrs.objsense == 1.0:
            self.results.problem.sense = minimize
        elif xprob_attrs.objsense == -1.0:
            self.results.problem.sense = maximize
        else:
            raise RuntimeError(
                'Unrecognized Xpress objective sense: {0}'.format(xprob_attrs.objsense)
            )

        try:
            soln.gap = (
                self.results.problem.upper_bound - self.results.problem.lower_bound
            )
        except TypeError:
            soln.gap = None

        self.results.problem.number_of_constraints = (
            xprob_attrs.rows + xprob_attrs.sets + xprob_attrs.qconstraints
        )
        self.results.problem.number_of_nonzeros = xprob_attrs.elems
        self.results.problem.number_of_variables = xprob_attrs.cols
        self.results.problem.number_of_integer_variables = xprob_attrs.mipents
        self.results.problem.number_of_continuous_variables = (
            xprob_attrs.cols - xprob_attrs.mipents
        )
        self.results.problem.number_of_objectives = 1
        self.results.problem.number_of_solutions = xprob_attrs.mipsols if is_mip else 1

        # if a solve was stopped by a limit, we still need to check to
        # see if there is a solution available - this may not always
        # be the case, both in LP and MIP contexts.
        if self._save_results:
            # This code in this if statement is only needed for backwards
            # compatibility. It is more efficient to set _save_results to
            # False and use load_vars, load_duals, etc.
            if have_soln:
                soln_variables = soln.variable
                soln_constraints = soln.constraint

                if extract_duals or extract_slacks:
                    xpress_cons = list(self._solver_con_to_pyomo_con_map.keys())
                    for con in xpress_cons:
                        soln_constraints[con.name] = {}

                xpress_vars = list(self._solver_var_to_pyomo_var_map.keys())
                try:
                    var_vals = xprob.getSolution(xpress_vars)
                    if extract_slacks:
                        slacks = self._getSlacks(xprob, xpress_cons)
                except xpress.ModelError:
                    # Xpress 9.5.0 has new behavior for unbounded
                    # problems that have mipsols > 0.  Previously
                    # getSolution() would return a solution, but now
                    # raises a ModelError (even though the deprecated
                    # getmipsol() will return a solution).  We will try
                    # to fall back on the [deprecated] getmipsol(), but
                    # if it fails, we will raise the original exception.
                    try:
                        var_vals = []
                        slacks = [] if extract_slacks else None
                        xprob.getmipsol(var_vals, slacks)
                        fail = 0
                    except:
                        fail = 1
                    if fail:
                        raise

                for xpress_var, val in zip(xpress_vars, var_vals):
                    pyomo_var = self._solver_var_to_pyomo_var_map[xpress_var]
                    if self._referenced_variables[pyomo_var] > 0:
                        soln_variables[xpress_var.name] = {"Value": val}

                if extract_reduced_costs:
                    vals = self._getRedCosts(xprob, xpress_vars)
                    for xpress_var, val in zip(xpress_vars, vals):
                        pyomo_var = self._solver_var_to_pyomo_var_map[xpress_var]
                        if self._referenced_variables[pyomo_var] > 0:
                            soln_variables[xpress_var.name]["Rc"] = val

                if extract_duals:
                    vals = self._getDuals(xprob, xpress_cons)
                    for val, con in zip(vals, xpress_cons):
                        soln_constraints[con.name]["Dual"] = val

                if extract_slacks:
                    for con, val in zip(xpress_cons, slacks):
                        if con in self._range_constraints:
                            ## for xpress, the slack on a range constraint
                            ## is based on the upper bound
                            lb = con.lb
                            ub = con.ub
                            ub_s = val
                            expr_val = ub - ub_s
                            lb_s = lb - expr_val
                            if abs(ub_s) > abs(lb_s):
                                soln_constraints[con.name]["Slack"] = ub_s
                            else:
                                soln_constraints[con.name]["Slack"] = lb_s
                        else:
                            soln_constraints[con.name]["Slack"] = val

        elif self._load_solutions:
            if have_soln:
                self.load_vars()

                if extract_reduced_costs:
                    self._load_rc()

                if extract_duals:
                    self._load_duals()

                if extract_slacks:
                    self._load_slacks()

        self.results.solution.insert(soln)

        # finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin.
        TempfileManager.pop(remove=not self._keepfiles)
        return DirectOrPersistentSolver._postsolve(self)

    def warm_start_capable(self):
        return True

    def _warm_start(self):
        mipsolval = list()
        mipsolcol = list()
        for pyomo_var, xpress_var in self._pyomo_var_to_solver_var_map.items():
            if pyomo_var.value is not None:
                mipsolval.append(value(pyomo_var))
                mipsolcol.append(xpress_var)
        if len(mipsolval) > 0:
            self._addMipSol(self._solver_model, mipsolval, mipsolcol)

    def _load_vars(self, vars_to_load=None):
        var_map = self._pyomo_var_to_solver_var_map
        ref_vars = self._referenced_variables
        if vars_to_load is None:
            vars_to_load = var_map.keys()

        xpress_vars_to_load = [var_map[pyomo_var] for pyomo_var in vars_to_load]
        vals = self._solver_model.getSolution(xpress_vars_to_load)

        for var, val in zip(vars_to_load, vals):
            if ref_vars[var] > 0:
                var.set_value(val, skip_validation=True)

    def _load_rc(self, vars_to_load=None):
        if not hasattr(self._pyomo_model, 'rc'):
            self._pyomo_model.rc = Suffix(direction=Suffix.IMPORT)
        var_map = self._pyomo_var_to_solver_var_map
        ref_vars = self._referenced_variables
        rc = self._pyomo_model.rc
        if vars_to_load is None:
            vars_to_load = var_map.keys()

        xpress_vars_to_load = [var_map[pyomo_var] for pyomo_var in vars_to_load]
        vals = self._getRedCosts(self._solver_model, xpress_vars_to_load)

        for var, val in zip(vars_to_load, vals):
            if ref_vars[var] > 0:
                rc[var] = val

    def _load_duals(self, cons_to_load=None):
        if not hasattr(self._pyomo_model, 'dual'):
            self._pyomo_model.dual = Suffix(direction=Suffix.IMPORT)
        con_map = self._pyomo_con_to_solver_con_map
        dual = self._pyomo_model.dual

        if cons_to_load is None:
            cons_to_load = con_map.keys()

        xpress_cons_to_load = [con_map[pyomo_con] for pyomo_con in cons_to_load]
        vals = self._getDuals(self._solver_model, xpress_cons_to_load)

        for pyomo_con, val in zip(cons_to_load, vals):
            dual[pyomo_con] = val

    def _load_slacks(self, cons_to_load=None):
        if not hasattr(self._pyomo_model, 'slack'):
            self._pyomo_model.slack = Suffix(direction=Suffix.IMPORT)
        con_map = self._pyomo_con_to_solver_con_map
        slack = self._pyomo_model.slack

        if cons_to_load is None:
            cons_to_load = con_map.keys()

        xpress_cons_to_load = [con_map[pyomo_con] for pyomo_con in cons_to_load]
        vals = self._getSlacks(self._solver_model, xpress_cons_to_load)

        for pyomo_con, xpress_con, val in zip(cons_to_load, xpress_cons_to_load, vals):
            if xpress_con in self._range_constraints:
                ## for xpress, the slack on a range constraint
                ## is based on the upper bound
                lb = xpress_con.lb
                ub = xpress_con.ub
                ub_s = val
                expr_val = ub - ub_s
                lb_s = lb - expr_val
                if abs(ub_s) > abs(lb_s):
                    slack[pyomo_con] = ub_s
                else:
                    slack[pyomo_con] = lb_s
            else:
                slack[pyomo_con] = val

    def load_duals(self, cons_to_load=None):
        """Load the duals into the 'dual' suffix. The 'dual' suffix must live
        on the parent model.

        Parameters
        ----------
        cons_to_load: list of Constraint

        """
        self._load_duals(cons_to_load)

    def load_rc(self, vars_to_load=None):
        """Load the reduced costs into the 'rc' suffix. The 'rc' suffix must
        live on the parent model.

        Parameters
        ----------
        vars_to_load: list of Var

        """
        self._load_rc(vars_to_load)

    def load_slacks(self, cons_to_load=None):
        """Load the values of the slack variables into the 'slack' suffix. The
        'slack' suffix must live on the parent model.

        Parameters
        ----------
        cons_to_load: list of Constraint

        """
        self._load_slacks(cons_to_load)


# Note: because _finalize_xpress_import references XpressDirect, we need
# to make sure to not attempt the xpress import until after the
# XpressDirect class is fully declared.
_xpress_importer = _xpress_importer_class()
xpress, xpress_available = attempt_import(
    'xpress',
    error_message=_xpress_importer,
    # Other forms of exceptions can be thrown by the xpress python
    # import.  For example, an xpress.InterfaceError exception is thrown
    # if the Xpress license is not valid.  Unfortunately, you can't
    # import without a license, which means we can't test for that
    # explicit exception!
    catch_exceptions=(Exception,),
    importer=_xpress_importer,
    callback=_finalize_xpress_import,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for indexed components
#

import pickle

import pyomo.common.unittest as unittest

from pyomo.environ import Var, Block, ConcreteModel, RangeSet, Set, Any
from pyomo.core.base.block import BlockData
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from pyomo.core.base.set import normalize_index


def _x_init(m, k):
    return k


def _y_init(m, i, j):
    return i * 10 + j


def _cx_init(b, k):
    i, j = b.index()[:2]
    return i * 100 + j * 10 + k


def _c(b, i, j):
    b.x = Var(b.model().K, initialize=_cx_init)


def _b(b, i, j):
    _c(b, i, j)
    b.c = Block(b.model().I, b.model().J, rule=_c)


def _bb(b, i, j, k):
    _c(b, i, j)
    b.c = Block(b.model().I, b.model().J, rule=_c)


class TestComponentSlices(unittest.TestCase):
    def setUp(self):
        self.m = m = ConcreteModel()
        m.I = RangeSet(1, 3)
        m.J = RangeSet(4, 6)
        m.K = RangeSet(7, 9)
        m.x = Var(m.K, initialize=_x_init)
        m.y = Var(m.I, m.J, initialize=_y_init)
        m.b = Block(m.I, m.J, rule=_b)
        m.bb = Block(m.I, m.J, m.K, rule=_bb)

    def tearDown(self):
        self.m = None

    def test_simple_getitem(self):
        self.assertIsInstance(self.m.b[1, 4], BlockData)

    def test_simple_getslice(self):
        _slicer = self.m.b[:, 4]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['b[1,4]', 'b[2,4]', 'b[3,4]'])

        _slicer = self.m.b[1, 4].c[:, 4]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['b[1,4].c[1,4]', 'b[1,4].c[2,4]', 'b[1,4].c[3,4]'])

    def test_wildcard_slice(self):
        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            _slicer = self.m.b[:]

        _slicer = self.m.b[...]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(
            ans,
            [
                'b[1,4]',
                'b[1,5]',
                'b[1,6]',
                'b[2,4]',
                'b[2,5]',
                'b[2,6]',
                'b[3,4]',
                'b[3,5]',
                'b[3,6]',
            ],
        )

        _slicer = self.m.b[1, ...]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['b[1,4]', 'b[1,5]', 'b[1,6]'])

        _slicer = self.m.b[..., 5]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['b[1,5]', 'b[2,5]', 'b[3,5]'])

        _slicer = self.m.bb[2, ..., 8]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['bb[2,4,8]', 'bb[2,5,8]', 'bb[2,6,8]'])

        _slicer = self.m.bb[:, ..., 8]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(
            ans,
            [
                'bb[1,4,8]',
                'bb[1,5,8]',
                'bb[1,6,8]',
                'bb[2,4,8]',
                'bb[2,5,8]',
                'bb[2,6,8]',
                'bb[3,4,8]',
                'bb[3,5,8]',
                'bb[3,6,8]',
            ],
        )

        _slicer = self.m.bb[:, :, ..., 8]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(
            ans,
            [
                'bb[1,4,8]',
                'bb[1,5,8]',
                'bb[1,6,8]',
                'bb[2,4,8]',
                'bb[2,5,8]',
                'bb[2,6,8]',
                'bb[3,4,8]',
                'bb[3,5,8]',
                'bb[3,6,8]',
            ],
        )

        _slicer = self.m.bb[:, ..., :, 8]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(
            ans,
            [
                'bb[1,4,8]',
                'bb[1,5,8]',
                'bb[1,6,8]',
                'bb[2,4,8]',
                'bb[2,5,8]',
                'bb[2,6,8]',
                'bb[3,4,8]',
                'bb[3,5,8]',
                'bb[3,6,8]',
            ],
        )

        _slicer = self.m.b[1, 4, ...]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['b[1,4]'])

        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            _slicer = self.m.b[1, 2, 3, ...]

        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            _slicer = self.m.b[1, :, 2]

        self.assertRaisesRegex(
            IndexError,
            'wildcard slice .* can only appear once',
            self.m.b.__getitem__,
            (Ellipsis, Ellipsis),
        )

    def test_any_slice(self):
        m = ConcreteModel()
        m.x = Var(Any, dense=False)
        m.x[1] = 1
        m.x[1, 1] = 2
        m.x[2] = 3
        self.assertEqual(list(str(_) for _ in m.x[:]), ['x[1]', 'x[2]'])
        self.assertEqual(list(str(_) for _ in m.x[:, :]), ['x[1,1]'])
        self.assertEqual(list(str(_) for _ in m.x[...]), ['x[1]', 'x[1,1]', 'x[2]'])

    def test_nonterminal_slice(self):
        _slicer = self.m.b[:, 4].x
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['b[1,4].x', 'b[2,4].x', 'b[3,4].x'])

        _slicer = self.m.b[:, 4].x[7]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, ['b[1,4].x[7]', 'b[2,4].x[7]', 'b[3,4].x[7]'])

    def test_nested_slices(self):
        _slicer = self.m.b[1, :].c[:, 4].x
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(
            ans,
            [
                'b[1,4].c[1,4].x',
                'b[1,4].c[2,4].x',
                'b[1,4].c[3,4].x',
                'b[1,5].c[1,4].x',
                'b[1,5].c[2,4].x',
                'b[1,5].c[3,4].x',
                'b[1,6].c[1,4].x',
                'b[1,6].c[2,4].x',
                'b[1,6].c[3,4].x',
            ],
        )

        _slicer = self.m.b[1, :].c[:, 4].x[8]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(
            ans,
            [
                'b[1,4].c[1,4].x[8]',
                'b[1,4].c[2,4].x[8]',
                'b[1,4].c[3,4].x[8]',
                'b[1,5].c[1,4].x[8]',
                'b[1,5].c[2,4].x[8]',
                'b[1,5].c[3,4].x[8]',
                'b[1,6].c[1,4].x[8]',
                'b[1,6].c[2,4].x[8]',
                'b[1,6].c[3,4].x[8]',
            ],
        )

    def test_component_function_slices(self):
        _slicer = self.m.component('b')[1, :].component('c')[:, 4].component('x')
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(
            ans,
            [
                'b[1,4].c[1,4].x',
                'b[1,4].c[2,4].x',
                'b[1,4].c[3,4].x',
                'b[1,5].c[1,4].x',
                'b[1,5].c[2,4].x',
                'b[1,5].c[3,4].x',
                'b[1,6].c[1,4].x',
                'b[1,6].c[2,4].x',
                'b[1,6].c[3,4].x',
            ],
        )

    def test_noncomponent_function_slices(self):
        ans = self.m.component('b')[1, :].component('c')[:, 4].x.fix(5)
        self.assertIsInstance(ans, list)
        self.assertEqual(ans, [None] * 9)

        ans = self.m.component('b')[1, :].component('c')[:, 4].x[:].is_fixed()
        self.assertIsInstance(ans, list)
        self.assertEqual(ans, [True] * (9 * 3))

        ans = self.m.component('b')[1, :].component('c')[:, 5].x[:].is_fixed()
        self.assertIsInstance(ans, list)
        self.assertEqual(ans, [False] * (9 * 3))

    def test_setattr_slices(self):
        init_sum = sum(self.m.b[:, :].c[:, :].x[:].value)
        init_vals = list(self.m.b[1, :].c[:, 4].x[:].value)
        self.m.b[1, :].c[:, 4].x[:].value = 0
        new_sum = sum(self.m.b[:, :].c[:, :].x[:].value)
        new_vals = list(self.m.b[1, :].c[:, 4].x[:].value)
        # nothing got deleted
        self.assertEqual(len(init_vals), len(new_vals))
        # the lists values were changes
        self.assertNotEqual(init_vals, new_vals)
        # the set values are all now zero
        self.assertEqual(sum(new_vals), 0)
        # nothing outside the set values changed
        self.assertEqual(init_sum - sum(init_vals), new_sum)

        # Test error on invalid attribute
        _slice = self.m.b[...].c[...].x[:]
        with self.assertRaisesRegex(
            AttributeError, ".*VarData' object has no attribute 'bogus'"
        ):
            _slice.bogus = 0
        # but disabling the exception flag will run without error
        _slice.attribute_errors_generate_exceptions = False
        # This doesn't do anything ... simply not raising an exception
        # is sufficient to verify the desired behavior
        _slice.bogus = 0

    def test_delattr_slices(self):
        self.m.b[1, :].c[:, 4].x.foo = 10
        # check that the attribute was added
        self.assertEqual(len(list(self.m.b[1, :].c[:, 4].x)), 3 * 3)
        self.assertEqual(sum(list(self.m.b[1, :].c[:, 4].x.foo)), 10 * 3 * 3)
        self.assertEqual(
            sum(list(1 if hasattr(x, 'foo') else 0 for x in self.m.b[:, :].c[:, :].x)),
            3 * 3,
        )

        _slice = self.m.b[1, :].c[:, 4].x.foo
        _slice._call_stack[-1] = (
            IndexedComponent_slice.del_attribute,
            _slice._call_stack[-1][1],
        )
        # call the iterator to delete the attributes
        list(_slice)
        self.assertEqual(
            sum(list(1 if hasattr(x, 'foo') else 0 for x in self.m.b[:, :].c[:, :].x)),
            0,
        )
        # calling the iterator again will raise an exception
        with self.assertRaisesRegex(AttributeError, 'foo'):
            list(_slice)
        # but disabling the exception flag will run without error
        _slice.attribute_errors_generate_exceptions = False
        # This doesn't do anything ... simply not raising an exception
        # is sufficient to verify the desired behavior
        list(_slice)

    def test_setitem_slices(self):
        init_sum = sum(self.m.b[:, :].c[:, :].x[:].value)
        init_vals = list(self.m.b[1, :].c[:, 4].x[:].value)
        self.m.b[1, :].c[:, 4].x[:] = 0
        new_sum = sum(self.m.b[:, :].c[:, :].x[:].value)
        new_vals = list(self.m.b[1, :].c[:, 4].x[:].value)
        # nothing got deleted
        self.assertEqual(len(init_vals), len(new_vals))
        # the lists values were changes
        self.assertNotEqual(init_vals, new_vals)
        # the set values are all now zero
        self.assertEqual(sum(new_vals), 0)
        # nothing outside the set values changed
        self.assertEqual(init_sum - sum(init_vals), new_sum)

        _slice = self.m.b[1, :].c[:, 4].x
        with self.assertRaisesRegex(
            KeyError,
            "Index 'bogus' is not valid for indexed "
            r"component 'b\[1,4\]\.c\[1,4\]\.x'",
        ):
            _slice['bogus'] = 0
        # but disabling the exception flag will run without error
        _slice.key_errors_generate_exceptions = False
        # This doesn't do anything ... simply not raising an exception
        # is sufficient to verify the desired behavior
        _slice['bogus'] = 0

    def test_setitem_component(self):
        init_sum = sum(self.m.x[:].value)
        init_vals = list(self.m.x[:].value)
        self.m.x[:] = 0
        new_sum = sum(self.m.x[:].value)
        new_vals = list(self.m.x[:].value)
        # nothing got deleted
        self.assertEqual(len(init_vals), len(new_vals))
        # the lists values were changes
        self.assertNotEqual(init_vals, new_vals)
        # the set values are all now zero
        self.assertEqual(sum(new_vals), 0)
        # nothing outside the set values changed
        self.assertEqual(init_sum - sum(init_vals), new_sum)

        init_sum = sum(self.m.y[:, :].value)
        init_vals = list(self.m.y[1, :].value)
        self.m.y[1, :] = 0
        new_sum = sum(self.m.y[:, :].value)
        new_vals = list(self.m.y[1, :].value)
        # nothing got deleted
        self.assertEqual(len(init_vals), len(new_vals))
        # the lists values were changes
        self.assertNotEqual(init_vals, new_vals)
        # the set values are all now zero
        self.assertEqual(sum(new_vals), 0)
        # nothing outside the set values changed
        self.assertEqual(init_sum - sum(init_vals), new_sum)

    def test_delitem_slices(self):
        init_all = list(self.m.b[:, :].c[:, :].x[:])
        init_tgt = list(self.m.b[1, :].c[:, 4].x[:])
        del self.m.b[1, :].c[:, 4].x[:]
        new_all = list(self.m.b[:, :].c[:, :].x[:])
        new_tgt = list(self.m.b[1, :].c[:, 4].x[:])

        self.assertEqual(len(init_tgt), 3 * 3 * 3)
        self.assertEqual(len(init_all), (3 * 3) * (3 * 3) * 3)
        self.assertEqual(len(new_tgt), 0)
        self.assertEqual(len(new_all), (3 * 3) * (3 * 3) * 3 - 3 * 3 * 3)

        _slice = self.m.b[2, :].c[:, 4].x
        with self.assertRaisesRegex(
            KeyError,
            "Index 'bogus' is not valid for indexed "
            r"component 'b\[2,4\]\.c\[1,4\]\.x'",
        ):
            del _slice['bogus']
        # but disabling the exception flag will run without error
        _slice.key_errors_generate_exceptions = False
        # This doesn't do anything ... simply not raising an exception
        # is sufficient to verify the desired behavior
        del _slice['bogus']
        # Nothing additional should have been deleted
        final_all = list(self.m.b[:, :].c[:, :].x[:])
        self.assertEqual(len(new_all), len(final_all))

    def test_delitem_component(self):
        init_all = list(self.m.bb[:, :, :])
        del self.m.bb[:, :, :]
        new_all = list(self.m.bb[:, :, :])
        self.assertEqual(len(init_all), 3 * 3 * 3)
        self.assertEqual(len(new_all), 0)

        init_all = list(self.m.b[:, :])
        init_tgt = list(self.m.b[1, :])
        del self.m.b[1, :]
        new_all = list(self.m.b[:, :])
        new_tgt = list(self.m.b[1, :])
        self.assertEqual(len(init_tgt), 3)
        self.assertEqual(len(init_all), 3 * 3)
        self.assertEqual(len(new_tgt), 0)
        self.assertEqual(len(new_all), 2 * 3)

    def test_empty_slices(self):
        _slicer = self.m.b[1, :].c[:, 1].x
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, [])

        _slicer = self.m.b[1, :].c[:, 4].x[1]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        _slicer.key_errors_generate_exceptions = False
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, [])

        _slicer = self.m.b[1, :].c[:, 4].y
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        _slicer.attribute_errors_generate_exceptions = False
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, [])

        _slicer = self.m.b[1, :].c[:, 4].component('y', False)
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        _slicer.call_errors_generate_exceptions = False
        ans = [str(x) for x in _slicer]
        self.assertEqual(ans, [])

        _slicer = self.m.b[1, :].c[:, 4].x[1]
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        _slicer.key_errors_generate_exceptions = True
        self.assertRaises(KeyError, _slicer.next)

        _slicer = self.m.b[1, :].c[:, 4].y
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        _slicer.attribute_errors_generate_exceptions = True
        self.assertRaises(AttributeError, _slicer.next)

        _slicer = self.m.b[1, :].c[:, 4].component('y', False)
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        _slicer.call_errors_generate_exceptions = True
        self.assertRaises(TypeError, _slicer.next)

        _slicer = self.m.b[1, :].c[:, 4].component()
        self.assertIsInstance(_slicer, IndexedComponent_slice)
        _slicer.call_errors_generate_exceptions = True
        self.assertRaises(TypeError, _slicer.next)

    def test_iterators(self):
        m = self.m

        _slice = self.m.x[...]
        self.assertEqual(list(_slice.wildcard_keys()), [7, 8, 9])
        self.assertEqual(
            list(_slice.wildcard_items()), [(7, m.x[7]), (8, m.x[8]), (9, m.x[9])]
        )
        self.assertEqual(list(_slice.expanded_keys()), [7, 8, 9])
        self.assertEqual(
            list(_slice.expanded_items()), [(7, m.x[7]), (8, m.x[8]), (9, m.x[9])]
        )

        _slice = self.m.b[...]
        self.assertEqual(
            list(_slice.wildcard_keys()),
            [(1, 4), (1, 5), (1, 6), (2, 4), (2, 5), (2, 6), (3, 4), (3, 5), (3, 6)],
        )
        self.assertEqual(
            list(_slice.wildcard_items()),
            [
                ((1, 4), m.b[1, 4]),
                ((1, 5), m.b[1, 5]),
                ((1, 6), m.b[1, 6]),
                ((2, 4), m.b[2, 4]),
                ((2, 5), m.b[2, 5]),
                ((2, 6), m.b[2, 6]),
                ((3, 4), m.b[3, 4]),
                ((3, 5), m.b[3, 5]),
                ((3, 6), m.b[3, 6]),
            ],
        )
        self.assertEqual(
            list(_slice.expanded_keys()),
            [(1, 4), (1, 5), (1, 6), (2, 4), (2, 5), (2, 6), (3, 4), (3, 5), (3, 6)],
        )
        self.assertEqual(
            list(_slice.expanded_items()),
            [
                ((1, 4), m.b[1, 4]),
                ((1, 5), m.b[1, 5]),
                ((1, 6), m.b[1, 6]),
                ((2, 4), m.b[2, 4]),
                ((2, 5), m.b[2, 5]),
                ((2, 6), m.b[2, 6]),
                ((3, 4), m.b[3, 4]),
                ((3, 5), m.b[3, 5]),
                ((3, 6), m.b[3, 6]),
            ],
        )

        _slice = self.m.b[1, :]
        self.assertEqual(list(_slice.wildcard_keys(False)), [4, 5, 6])
        self.assertEqual(
            list(_slice.wildcard_items(False)),
            [(4, m.b[1, 4]), (5, m.b[1, 5]), (6, m.b[1, 6])],
        )
        self.assertEqual(list(_slice.expanded_keys()), [(1, 4), (1, 5), (1, 6)])
        self.assertEqual(
            list(_slice.expanded_items()),
            [((1, 4), m.b[1, 4]), ((1, 5), m.b[1, 5]), ((1, 6), m.b[1, 6])],
        )

    def test_pickle_slices(self):
        m = self.m
        _slicer = m.b[1, :].c[:, 4].x
        _new_slicer = pickle.loads(pickle.dumps(_slicer))

        self.assertIsNot(_slicer, _new_slicer)
        self.assertIsNot(_slicer._call_stack, _new_slicer._call_stack)
        self.assertIs(type(_slicer._call_stack), type(_new_slicer._call_stack))
        self.assertEqual(len(_slicer._call_stack), len(_new_slicer._call_stack))

        ref = [
            'b[1,4].c[1,4].x',
            'b[1,4].c[2,4].x',
            'b[1,4].c[3,4].x',
            'b[1,5].c[1,4].x',
            'b[1,5].c[2,4].x',
            'b[1,5].c[3,4].x',
            'b[1,6].c[1,4].x',
            'b[1,6].c[2,4].x',
            'b[1,6].c[3,4].x',
        ]
        self.assertEqual([str(x) for x in _slicer], ref)
        self.assertEqual([str(x) for x in _new_slicer], ref)
        for x, y in zip(iter(_slicer), iter(_new_slicer)):
            self.assertIs(type(x), type(y))
            self.assertEqual(x.name, y.name)
            self.assertIsNot(x, y)

    def test_clone_on_model(self):
        m = self.m
        m.slicer = m.b[1, :].c[:, 4].x
        n = m.clone()

        self.assertIsNot(m, n)
        self.assertIsNot(m.slicer, n.slicer)
        self.assertIsNot(m.slicer._call_stack, n.slicer._call_stack)
        self.assertIs(type(m.slicer._call_stack), type(n.slicer._call_stack))
        self.assertEqual(len(m.slicer._call_stack), len(n.slicer._call_stack))

        ref = [
            'b[1,4].c[1,4].x',
            'b[1,4].c[2,4].x',
            'b[1,4].c[3,4].x',
            'b[1,5].c[1,4].x',
            'b[1,5].c[2,4].x',
            'b[1,5].c[3,4].x',
            'b[1,6].c[1,4].x',
            'b[1,6].c[2,4].x',
            'b[1,6].c[3,4].x',
        ]
        self.assertEqual([str(x) for x in m.slicer], ref)
        self.assertEqual([str(x) for x in n.slicer], ref)
        for x, y in zip(iter(m.slicer), iter(n.slicer)):
            self.assertIs(type(x), type(y))
            self.assertEqual(x.name, y.name)
            self.assertIsNot(x, y)
            self.assertIs(x.model(), m)
            self.assertIs(y.model(), n)

    def test_hash_eqality(self):
        m = self.m
        a = m.b[1, :].c[:, ..., 4].x
        b = m.b[1, :].c[1, ..., :].x
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, m)

        self.assertEqual(a, a)
        self.assertEqual(a, m.b[1, :].c[:, ..., 4].x)

        _set = set([a, b])
        self.assertEqual(len(_set), 2)
        _set.add(m.b[1, :].c[:, ..., 4].x)
        self.assertEqual(len(_set), 2)
        _set.add(m.b[1, :].c[:, 4].x)
        self.assertEqual(len(_set), 3)

    def test_duplicate(self):
        m = self.m
        a = m.b[1, :].c[:, ..., 4]

        b = a.x
        self.assertIs(a._call_stack, b._call_stack)
        self.assertEqual(a._len + 1, b._len)

        c = a.y
        self.assertEqual(a._len + 1, c._len)
        self.assertIsNot(a._call_stack, c._call_stack)

        b1 = b.duplicate()
        self.assertIsNot(a._call_stack, b1._call_stack)
        self.assertEqual(a._len + 1, b1._len)
        self.assertEqual(hash(b), hash(b1))

    def test_invalid_slices(self):
        m = self.m
        m.x = Var()
        for var in m.x[:]:
            self.assertIs(var, m.x)

        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            _slicer = m.b[:]

        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            _slicer = m.b[:, :, :]

        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            _slicer = m.b[:, :, :, ...]

        # valid slice for b, but not c
        _slicer = m.b[:, :, ...].c[:, :, :].x
        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            # Error not raised immediately because accessing c is deferred
            # until iteration.
            list(_slicer)

        # valid slice for b, but not c
        _slicer = m.b[2, :].c[:].x
        with self.assertRaisesRegex(
            IndexError,
            'Index .* contains an invalid number of entries for component .*',
        ):
            list(_slicer)

    def test_nondim_set(self):
        m = ConcreteModel()
        m.I = Set(dimen=None, initialize=[1, (2, 3)])
        m.x = Var(m.I)

        ref = list(m.x[:])
        self.assertEqual(len(ref), 1)
        self.assertIs(ref[0], m.x[1])

        ref = list(m.x[:, ..., :])
        self.assertEqual(len(ref), 1)
        self.assertIs(ref[0], m.x[2, 3])

        ref = list(m.x[2, ...])
        self.assertEqual(len(ref), 1)
        self.assertIs(ref[0], m.x[2, 3])

        _old_flatten = normalize_index.flatten
        try:
            normalize_index.flatten = False

            m = ConcreteModel()
            m.I = Set(dimen=None, initialize=[1, (2, 3)])
            m.x = Var(m.I)

            ref = list(m.x[:])
            self.assertEqual(len(ref), 2)
            self.assertIs(ref[0], m.x[1])

            with self.assertRaisesRegex(
                IndexError,
                'Index .* contains an invalid number of entries for component .*',
            ):
                # If we are not flattening the sets, then
                # non-dimensioned Sets *still expect a single "slice".
                list(m.x[:, ..., :])

        finally:
            normalize_index.flatten = _old_flatten

    def test_UnknownSetDimen(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2, 3])
        m.J = Set()
        m.x = Var(m.I, m.J)

        with self.assertRaisesRegex(
            IndexError,
            'Slicing components relies on knowing the underlying set dimensionality',
        ):
            ref = list(m.x[:, :])

    def test_flatten_false(self):
        _old_flatten = normalize_index.flatten
        try:
            normalize_index.flatten = False

            m = ConcreteModel()
            m.I = Set(initialize=range(2))
            m.J = Set(initialize=range(2, 4))
            m.K = Set(initialize=['a', 'b', 'c'])
            m.IJ = m.I * m.J

            m.a = Var(m.I, m.J, m.K)
            m.b = Var(m.IJ, m.K)
            m.c = Var()

            with self.assertRaisesRegex(
                IndexError,
                'Index .* contains an invalid number of entries for component .*',
            ):
                _slicer = m.a[(0, 2), :]

            _slicer = m.a[0, 2, :]
            names = ['a[0,2,a]', 'a[0,2,b]', 'a[0,2,c]']
            self.assertEqual(names, [var.name for var in _slicer])

            with self.assertRaisesRegex(
                IndexError,
                'Index .* contains an invalid number of entries for component .*',
            ):
                _slicer = m.b[0, 2, :]

            _slicer = m.b[(0, 2), :]
            names = ['b[(0,2),a]', 'b[(0,2),b]', 'b[(0,2),c]']
            self.assertEqual(names, [var.name for var in _slicer])

            with self.assertRaisesRegex(
                IndexError,
                'Index .* contains an invalid number of entries for component .*',
            ):
                _slicer = m.b[:, 2, 'b']

            _slicer = m.b[:, 'b']
            names = ['b[(0,2),b]', 'b[(0,3),b]', 'b[(1,2),b]', 'b[(1,3),b]']
            self.assertEqual(names, [var.name for var in _slicer])
            _slicer = m.b[..., 'b']
            self.assertEqual(names, [var.name for var in _slicer])

            _slicer = m.b[0, ...]
            self.assertEqual([], [var.name for var in _slicer])

            _slicer = m.c[:]
            self.assertEqual(['c'], [var.name for var in _slicer])

        finally:
            normalize_index.flatten = _old_flatten

    def test_compare_1dim_slice(self):
        m = ConcreteModel()
        m.I = Set(initialize=range(2))
        m.J = Set(initialize=range(2, 4))
        m.K = Set(initialize=['a', 'b'])

        @m.Block(m.I, m.J)
        def b(b, i, j):
            b.v = Var(m.K)

        self.assertEqual(m.b[0, :].v[:], m.b[0, :].v[:])
        self.assertNotEqual(m.b[0, :].v[:], m.b[0, :].v['a'])

    def test_str(self):
        m = ConcreteModel()
        m.b = Block()
        # Note that we are testing the string representation of a slice,
        # not if the slice is valid
        s = m.b[...].x[:, 1:2, 1:5:2, ::1, 5, 'a'].component('foo', kwarg=1)
        self.assertEqual(
            str(s), "b[...].x[:, 1:2, 1:5:2, ::1, 5, 'a'].component('foo', kwarg=1)"
        )

        # To test set / del, we want to form the IndexedComponent_slice
        # without evaluating it
        s = m.b[...]
        self.assertEqual(
            str(
                IndexedComponent_slice(
                    s, (IndexedComponent_slice.del_attribute, 'bogus')
                )
            ),
            'del b[...].bogus',
        )
        self.assertEqual(
            str(
                IndexedComponent_slice(
                    s, (IndexedComponent_slice.set_attribute, 'bogus', 10)
                )
            ),
            'b[...].bogus = 10',
        )

    def test_slice_to_componentdata(self):
        m = ConcreteModel()
        m.x = Var([1, 2])
        i = IndexedComponent_slice(m.x, None, None, None)[2]
        self.assertEqual(list(i), [m.x[2]])


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Configuration options for incidence graph generation"""

import enum
from pyomo.common.config import ConfigDict, ConfigValue, InEnum
from pyomo.common.modeling import NOTSET
from pyomo.repn.ampl import AMPLRepnVisitor
from pyomo.repn.util import FileDeterminism, FileDeterminism_to_SortComponents


class IncidenceMethod(enum.Enum):
    """Methods for identifying variables that participate in expressions"""

    identify_variables = 0
    """Use ``pyomo.core.expr.visitor.identify_variables``"""

    standard_repn = 1
    """Use ``pyomo.repn.standard_repn.generate_standard_repn``"""

    standard_repn_compute_values = 2
    """Use ``pyomo.repn.standard_repn.generate_standard_repn`` with
    ``compute_values=True``
    """

    ampl_repn = 3
    """Use ``pyomo.repn.ampl.AMPLRepnVisitor``"""


class IncidenceOrder(enum.Enum):

    dulmage_mendelsohn_upper = 0

    dulmage_mendelsohn_lower = 1


_include_fixed = ConfigValue(
    default=False,
    domain=bool,
    description="Include fixed variables",
    doc=(
        "Flag indicating whether fixed variables should be included in the"
        " incidence graph"
    ),
)


_linear_only = ConfigValue(
    default=False,
    domain=bool,
    description="Identify only variables that participate linearly",
    doc=(
        "Flag indicating whether only variables that participate linearly should"
        " be included."
    ),
)


_method = ConfigValue(
    default=IncidenceMethod.standard_repn,
    domain=InEnum(IncidenceMethod),
    description="Method used to identify incident variables",
)


def _amplrepnvisitor_validator(visitor):
    if not isinstance(visitor, AMPLRepnVisitor):
        raise TypeError(
            "'visitor' config argument should be an instance of AMPLRepnVisitor"
        )
    return visitor


_ampl_repn_visitor = ConfigValue(
    default=None,
    domain=_amplrepnvisitor_validator,
    description="Visitor used to generate AMPLRepn of each constraint",
)


IncidenceConfig = ConfigDict()
"""Options for incidence graph generation

- ``include_fixed`` -- Flag indicating whether fixed variables should be included
  in the incidence graph
- ``linear_only`` -- Flag indicating whether only variables that participate linearly
  should be included.
- ``method`` -- Method used to identify incident variables. Must be a value of the
  ``IncidenceMethod`` enum.
- ``_ampl_repn_visitor`` -- Expression visitor used to generate ``AMPLRepn`` of each
  constraint. Must be an instance of ``AMPLRepnVisitor``. *This option is constructed
  automatically when needed and should not be set by users!*

"""


IncidenceConfig.declare("include_fixed", _include_fixed)


IncidenceConfig.declare("linear_only", _linear_only)


IncidenceConfig.declare("method", _method)


IncidenceConfig.declare("_ampl_repn_visitor", _ampl_repn_visitor)


def get_config_from_kwds(**kwds):
    """Get an instance of IncidenceConfig from provided keyword arguments.

    If the ``method`` argument is ``IncidenceMethod.ampl_repn`` and no
    ``AMPLRepnVisitor`` has been provided, a new ``AMPLRepnVisitor`` is
    constructed. This function should generally be used by callers such
    as ``IncidenceGraphInterface`` to ensure that a visitor is created then
    re-used when calling ``get_incident_variables`` in a loop.

    """
    if (
        kwds.get("method", None) is IncidenceMethod.ampl_repn
        and kwds.get("_ampl_repn_visitor", None) is None
    ):
        subexpression_cache = {}
        external_functions = {}
        var_map = {}
        used_named_expressions = set()
        symbolic_solver_labels = False
        # TODO: Explore potential performance benefit of exporting defined variables.
        # This likely only shows up if we can preserve the subexpression cache across
        # multiple constraint expressions.
        export_defined_variables = False
        sorter = FileDeterminism_to_SortComponents(FileDeterminism.ORDERED)
        amplvisitor = AMPLRepnVisitor(
            subexpression_cache,
            external_functions,
            var_map,
            used_named_expressions,
            symbolic_solver_labels,
            export_defined_variables,
            sorter,
        )
        kwds["_ampl_repn_visitor"] = amplvisitor
    return IncidenceConfig(kwds)
